# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import os, sys, time, importlib
import subprocess
import numpy as np
from matplotlib import pyplot as plt
import cPickle as pkl
import gzip
import h5py
import getpass
import argparse

user_name = getpass.getuser()


def import_variable_from_file(file_path, class_name):
    directory = os.path.dirname(file_path)
    sys.path.append(directory)
    mod_name = os.path.split(file_path)[1]
    if mod_name[-3:] == '.py':
        mod_name = mod_name[:-3]

    mod = importlib.import_module(mod_name)
    sys.path.pop(-1)
    cls = getattr(mod, class_name)
    return cls


def parseargs(gpu):
    def convert(s):
        if s=='auto':
            return s
        elif s.lower()=='false':
            return False
        elif s.lower()==None:
            return None
        else:
            return int(s)
    
    parser = argparse.ArgumentParser(
    usage="elektronn-train </path/to_config_file> [--gpu={Auto|False|<int>}]")
    
    parser.add_argument("--gpu", default=gpu, type=convert, choices=['auto', False, None]+range(0,100))
    parser.add_argument("config", type=str)
    parsed = parser.parse_args()
    return parsed.config, parsed.gpu

    

def parseargs_dev(args, config_file, gpu):
    """
    Parses the commandline arguments if ``elektronn-train`` is called as:

    "elektronn-train [config=</path/to_file>] [ gpu={Auto|False|<int>}]"
    """
    for arg in args:
        if arg.startswith('gpu='):
            arg = arg.replace('gpu=', '')
            if (arg == 'False'):
                gpu = False
                print "Not assigning a GPU explicitly (theano.config default is used)"

            else:
                if arg in ['auto', 'Auto']:
                    gpu = get_free_gpu()
                    print "Automatically assigned free gpu%i" % gpu
                else:
                    gpu = int(arg)

                try:
                    gpu = int(gpu)
                except:
                    sys.excepthook(*sys.exc_info())
                    raise ValueError("Invalid GPU argument %s (maybe no free GPU?)" % str(gpu))

        elif arg.startswith('config='):
            # Override above default configuration parameters with config-file
            arg = arg.replace('config=', '')
            config_file = arg
    return config_file, gpu


def _check_if_gpu_is_free(nb_gpu):
    process_output = subprocess.Popen('nvidia-smi -i %d -q -d PIDS' % nb_gpu,
                                      stdout=subprocess.PIPE,
                                      shell=True).communicate()[0]
    if "Process ID" in process_output and "Used GPU Memory" in process_output:
        return 0
    else:
        return 1


def _get_number_gpus():
    process_output = subprocess.Popen('nvidia-smi -L',
                                      stdout=subprocess.PIPE,
                                      shell=True).communicate()[0]
    nb_gpus = 0
    while True:
        if "GPU %d" % nb_gpus in process_output:
            nb_gpus += 1
        else:
            break
    return nb_gpus


def get_free_gpu(wait=0, nb_gpus=-1):
    if nb_gpus == -1:
        nb_gpus = _get_number_gpus()
    while True:
        for nb_gpu in range(nb_gpus):
            if _check_if_gpu_is_free(nb_gpu) == 1:
                return nb_gpu
        if wait > 0:
            time.sleep(2)
        else:
            return -1


def initGPU(gpu):
    if gpu is not False:
        import theano.sandbox.cuda  # import theano AFTER backing up scripts (because it takes a while)
        if theano.sandbox.cuda.cuda_available:
            print "Initialising GPU to %s" % gpu
            try:
                if gpu is None:
                    theano.sandbox.cuda.use("gpu0")
                else:
                    theano.sandbox.cuda.use("gpu" + str(gpu))

            except:
                sys.excepthook(*sys.exc_info())
        else:
            print "'gpu' is not 'False' but CUDA is not available. Falling back to cpu"


def xyz2zyx(shapes):
    """
    Swaps dimension order for list of (filter) shapes.
    This is needed to allow users to specify 2d and 3d filters in the same order.
    """
    if not hasattr(shapes[0], '__len__'):
        return shapes
    else:
        return map(lambda s: [s[2], s[0], s[1]], shapes)

### Printing/Plotting #########################################################################################


def _SMA(c, n):
    """
    Returns box-SMA of c with box length n, the returned array has the same
    length as c and is const-padded at the beginning
    """
    if n == 0: return c
    ret = np.cumsum(c, dtype=float)
    ret[n:] = (ret[n:] - ret[:-n]) / n
    m = min(n, len(c))
    ret[:n] = ret[:n] / np.arange(1, m + 1)  # unsmoothed
    return ret


def plotInfoFromFiles(path, save_name, autoscale=True):
    """
    Create the plots from backup files in the CNN directory (e.g. if plotting was not on during training).
    The plots are generated as pngs in the current working directory and will not show up.

    Parameters
    ----------

    path: string
      Path to CNN-folder
    save_name: string
      name of cnn / file prefix
    autoscale: Bool
      If true axis are optimised for value read-off, if false, mpl default scaling is used
    """
    os.chdir(path)
    timeline = np.load('Backup/' + save_name + ".timeline.npy")
    if timeline.shape[1] == 4:
        ix = np.arange(len(timeline))
        timeline = np.hstack((ix[:, None], timeline))

    history = np.load('Backup/' + save_name + ".history.npy")
    if history.shape[1] == 6:
        ix = np.arange(len(history))
        history = np.hstack((ix[:, None], history))

    try:
        errors = np.load('Backup/' + save_name + ".errors.npy")
    except:
        try:
            errors = np.load('Backup/' + save_name + ".ErrorTimeline.npy")
        except:
            errors = [[0, 0, 0]]
    if errors.shape[1] == 3:
        ix = np.arange(len(errors))
        errors = np.hstack((ix[:, None], errors))

    try:
        CG_timeline = np.load('Backup/' + save_name + ".CG_timeline.npy")
    except:
        CG_timeline = []

    plotInfo(timeline, history, CG_timeline, errors, save_name, autoscale)


def saveHist(timeline, history, CG_timeline, errors, save_name):
    np.save('Backup/' + save_name + ".history", history)
    np.save('Backup/' + save_name + ".timeline", timeline)
    np.save('Backup/' + save_name + ".errors", errors)
    np.save('Backup/' + save_name + ".CGtimeline", CG_timeline)


def plotInfo(timeline,
             history,
             CG_timeline,
             errors,
             save_name,
             autoscale=True):
    """Plot graphical info during Training"""
    plt.ioff()

    def addStepTicks(ax, times, steps, num=5):
        N = int(steps[-1])
        k = max(N / num, 1)
        k = int(np.log10(k))  # 10-base of locators
        m = int(np.round(float(N) / (num * 10**k)))  # multiple of base
        s = m * 10**k
        x_labs = np.arange(0, N, s, dtype=np.int)
        x_ticks = np.interp(x_labs, steps, times)
        ax.set_xticks(x_ticks)
        ax.set_xticklabels(x_labs)
        ax.set_xlim(0, times[-1])
        ax.set_xlabel('Update steps (%s)' % ("{0:,d}".format(N)))

    try:
        timeline = np.array(timeline)
        history = np.array(history)
        CG_timeline = np.array(CG_timeline)
        errors = np.array(errors)
        saveHist(timeline, history, CG_timeline, errors, save_name)

        # Subsample points for plotting
        s = max((len(timeline) // 2000), 1)
        timeline = timeline[::s]
        s = max((len(history) // 2000), 1)
        history = history[::s]
        s = max((len(CG_timeline) // 2000), 1)
        CG_timeline = CG_timeline[::s]
        s = max((len(errors) // 2000), 1)
        errors = errors[::s]

        # secs --> mins
        timeline[:, 1] = timeline[:, 1] / 60
        history[:, 1] = history[:, 1] / 60
        errors[:, 1] = errors[:, 1] / 60
        runtime = str(int(timeline[-1, 1])) + ' mins' if (
            timeline[-1, 1] < 120) else str(int(timeline[-1, 1] / 60)) + ' hrs'

        if not np.any(np.isnan(
                history[:, 5])):  # check if valid data is available
            nll_cap = history[0, 5] if len(history) < 3 else max(
                history[2, 5], history[2, 2]) + 0.2
        else:
            nll_cap = history[0, 4] if len(history) < 3 else max(
                history[2, 4], history[2, 2]) + 0.2

        #---------------------------------------------------------------------------------------------------------
        # Timeline, NLLs
        plt.figure(figsize=(16, 12))
        plt.subplot(211)
        plt.plot(timeline[:, 1], timeline[:, 3], 'b-', label='Update NLL')
        plt.plot(timeline[:, 1], timeline[:, 2], 'k-', label='Smooth update NLL', linewidth=3)

        if autoscale:
            plt.ylim(0, nll_cap)
            plt.xlim(0, timeline[-1, 1])  # 105% of time (leave some white to the right)

        plt.legend(loc=0)
        nll = timeline[-30:, 2].mean()
        plt.hlines(nll, 0, timeline[-1:, 0])  # last smooth NLL
        plt.xlabel("time [min] (%s)" % runtime)

        ax = plt.twiny()
        addStepTicks(ax, timeline[:, 1], timeline[:, 0])

        # NLL vs Prevalence
        plt.subplot(212)
        plt.scatter(timeline[:, 4], timeline[:, 3], c=timeline[:, 0], cmap='binary', edgecolors='none')
        if autoscale:
            plt.ylim(0, 1)
            plt.xlim(-0.01, 1)

        plt.xlabel('Mean label of batch')
        plt.ylabel('NLL')

        plt.savefig(save_name + ".timeline.png", bbox_inches='tight')

        #---------------------------------------------------------------------------------------------------------
        # History NLL
        plt.figure(figsize=(16, 12))
        plt.subplot(211)
        plt.plot(history[:, 1], history[:, 2], 'b-', label='Smooth update NLL', linewidth=3)
        #plt.plot(history[:,1], history[:,3], 'k-', label = 'Update NLL')
        plt.plot(history[:, 1], history[:, 4], 'g-', label='Train NLL', linewidth=3)
        plt.plot(history[:, 1], history[:, 5], 'r-', label='Valid NLL', linewidth=3)
        if autoscale:
            plt.ylim(0, nll_cap)
            plt.xlim(0, history[-1, 1])

        plt.legend(loc=0)
        plt.xlabel("time [min] (%s)" % runtime)

        ax = plt.twiny()
        addStepTicks(ax, history[:, 1], history[:, 0])

        # History NLL gains
        plt.subplot(212)
        plt.plot(history[:, 1], history[:, 6], 'b-', label='NLL Gain at update')

        plt.hlines(0, 0, history[-1:, 1], linestyles='dotted')
        plt.xlabel("time [min] (%s)" % runtime)

        plt.plot(history[:, 1], history[:, -1], 'r-', label='LR')
        plt.legend(loc=0)

        std = history[3:, 6].std() * 2 if len(history) > 3 else 1.0
        if autoscale:
            plt.ylim(-std, std + 1e-10)  # add epsilon to suppress matplotlib warning in case of CG
            plt.xlim(0, history[-1, 1])

        plt.savefig(save_name + ".history.png", bbox_inches='tight')

        #---------------------------------------------------------------------------------------------------------
        if CG_timeline.size > 0:  # [nll, t, coeff, count]
            "[nll, t, coeff, count]"
            plt.figure(figsize=(16, 12))
            plt.subplot(211)
            plt.plot(CG_timeline[:, 0], label='CG Update NLL')
            if autoscale:
                plt.ylim(0, nll_cap)
                plt.xlim(0, len(CG_timeline))

            plt.grid()
            plt.legend(loc=0)
            plt.xlabel("Update steps")

            plt.subplot(212)
            noise = np.random.rand(len(CG_timeline)) * 0.001 - 0.0005
            plt.scatter(CG_timeline[:, 2], CG_timeline[:, 1] + noise)
            plt.xlabel('coefficient')
            plt.ylabel('accepted step')
            step = CG_timeline[:, 1].mean()
            std = CG_timeline[:, 1].std()
            count = CG_timeline[:, 3].mean()
            plt.title("Step size vs. conjugacy coeff, avg_step=%.6f, step_std=%8f, avg_count=%.1f" % (step, std, count))
            plt.savefig(save_name + ".CGtimeline.png", bbox_inches='tight')

        #---------------------------------------------------------------------------------------------------------
        if errors.size > 0:  # [t, coeff, count]
            cutoff = 2
            if len(errors) > (cutoff + 1):
                errors = errors[cutoff:]
            if not np.any(np.isnan(
                    errors[:, 3])):  # check if valid data is available
                err_cap = errors[:, 3].mean() * 3
            else:
                err_cap = errors[:, 2].mean() * 3

            plt.figure(figsize=(16, 6))
            plt.plot(errors[:, 1], errors[:, 2], 'g--', label='Train error', linewidth=1)
            plt.plot(errors[:, 1], errors[:, 3], 'r--', label='Valid Error', linewidth=1)
            plt.plot(errors[:, 1], _SMA(errors[:, 2], 8), 'g-', label='Smooth train error', linewidth=3)
            if not np.any(np.isnan(_SMA(errors[:, 3], 8))):
                plt.plot(errors[:, 1], _SMA(errors[:, 3], 8), 'r-', label='Smooth valid Error', linewidth=3)
            if autoscale:
                plt.ylim(0, err_cap)
                plt.xlim(0, errors[-1, 1])

            plt.grid()
            plt.legend(loc=0)
            plt.xlabel("time [min] (%s)" % runtime)

            ax = plt.twiny()
            addStepTicks(ax, errors[:, 1], errors[:, 0])

            plt.savefig(save_name + ".Errors.png", bbox_inches='tight')

        plt.close('all')
    except ValueError:
        # When arrays are empty
        print "An error occoured durint plotting"


def previewDiffPlot(names, root_dir='~/CNN_Training/3D', block_name=0, c=1, z=0, number=1, save=True):
    """
    Visualisation tool to compare the predictions of 2 or multiple CNNs
    It is assumed that

    Parameters
    ----------

    names: list of str
      Folder/Save names of the CNNs
    root_dir: str
      path in which the CNN folders are located
    block_name: int/str
      Number/Name of the prediction preview example ("...pred_<i>_c..")

    """

    def getDiff(p1, p2):
        plt.ion()
        p1 = plt.imread(p1).astype(np.float32) / 2**16
        p2 = plt.imread(p2).astype(np.float32) / 2**16
        out_sh = np.minimum(p1.shape, p2.shape)

        if np.any(p1.shape != out_sh):
            pad_s = np.subtract(p1.shape, out_sh, ) // 2
            p1 = p1[pad_s[0]:-pad_s[0], pad_s[1]:-pad_s[1]]
        if np.any(p2.shape != out_sh):
            pad_s = np.subtract(p2.shape, out_sh) // 2
            p2 = p2[pad_s[0]:-pad_s[0], pad_s[1]:-pad_s[1]]

        g = p1
        b = p2
        r = np.minimum(p1, p2)
        img = np.dstack([r, g, b])  # color image
        return img

    root_dir = os.path.expanduser(root_dir)
    os.chdir(root_dir)
    block_name = str(block_name)
    N = (len(names) * (len(names) - 1)) / 2
    n = int(np.sqrt(N * 16.0 / 9))  # aim: ratio 16:9
    m = int(N * 1.0 / n + 1)
    if N == 1:
        m = 1

    plt.figure(figsize=(16, 9))
    count = 1
    for i in xrange(len(names)):
        for j in xrange(i + 1, len(names)):
            p1 = "%s/%s/%s-pred-%s-c%i-z%i-%ihrs.png" % (root_dir, names[i], names[i], block_name, c, z, number)
            p2 = "%s/%s/%s-pred-%s-c%i-z%i-%ihrs.png" % (root_dir, names[j], names[j], block_name, c, z, number)
            diff = getDiff(p1, p2)

            plt.subplot(n, m, count)
            count += 1
            plt.imshow(diff, interpolation='none')
            plt.title("%s (green) vs. %s (blue)" % (names[i], names[j]))
            if save:
                plt.imsave("%s_vs_%s_pred-%s-c%i-z%i-%ihrs.png" % (names[i], names[j], block_name, c, z, number), diff)

    plt.show()
    if save:
        plt.savefig("Compare_pred-%s-c%i-z%i-%ihrs.png" % (block_name, c, z, number), bbox_inches='tight')
    plt.ioff()


def pprintmenu(save_name):
    """print menu string"""
    s = """ELEKTRONN MENU\n==============\n
    >> %s <<\n\
    Shortcuts:\n\
    'q' (leave menu),\t\t\
    'abort' (saving params),\n\
    'kill'(no saving),\t\t\
    'save'/'load' (opt:filename),\n\
    'sf'/' (show filters)',\t\
    'smooth' (smooth filters),\n\
    'sethist <int>',\t\t\
    'setlr <float>',\n\
    'setmom <float>' ,\t\t\
    'params' print info,\n\
    Change Training Optimizer :('SGD','CG', 'RPROP', 'LBFGS')\n\
    For everything else enter a command in the command line\n""" % save_name
    print s


def userInput(cnn, history_freq):
    user_input_string = raw_input("%s@ELEKTRONN: " % (user_name))

    words = user_input_string.split(" ")

    if words[0] == "q":
        return "q"
    elif words[0] == "abort":
        return "abort"
    elif words[0] == "kill":
        return "kill"
    elif words[0] in ["SGD", "CG", 'RPROP', 'LBFGS', 'Adam']:
        print "Changing Training mode..."
        return words[0]
    elif words[0] == "save":
        if (len(words) == 1):
            cnn.saveParameters()
        else:
            cnn.saveParameters(words[1])
        print "Saving done"
    elif words[0] == "load":
        if len(words) == 1:
            cnn.loadParameters()
        else:
            cnn.loadParameters(words[1])
        print "Loaded!"

    elif words[0] == "smooth":
        print "Filter Smoothing not implemented"
    elif words[0] == "sethist":
        if words[1].isdigit():
            history_freq[0] = int(words[1])
        else:
            print "Could not convert to number"
    elif words[0] == "setlr":
        cnn.setSGDLR(np.float32(float(words[1])))
    elif words[0] == "setmom":
        cnn.setSGDMomentum(np.float32(float(words[1])))
    elif words[0] == "params":
        n = len(cnn.layers)
        for j in range(len(cnn.params)):
            pa = cnn.params[j].get_value()
            print 'Layer  ' + str(n - int(j / 2)) + ': filter_shape =', np.shape(pa)
            print 'Layer {0:2d}: np.mean(abs)={1:2.6e}, np.std={1:2.6e} np.median(abs)={2:2.6e}  '.format(
                n - int(j / 2), np.mean(abs(pa)), np.std(pa), np.median(abs(pa)))

    else:
        return user_input_string  # this string will be executed in main

##############################################################################################################


def pickleSave(data, file_name):
    """
    Writes one or many objects to pickle file

    data:
      single objects to save or iterable of objects to save.
      For iterable, all objects are written in this order to the file.
    file_name: string
      path/name of destination file
    """
    with open(file_name, 'wb') as f:
        if not hasattr(data, '__len__'):
            pkl.dump(data, f, protocol=2)
        else:
            for d in data:
                pkl.dump(d, f, protocol=2)


def pickleLoad(file_name):
    """
    Loads all object that are saved in the pickle file.
    Multiple objects are returned as list.
    """
    ret = []
    try:
        with open(file_name, 'rb') as f:
            try:
                while True:
                    # Python 3 needs explicit encoding specification,
                    # which Python 2 lacks:
                    if sys.version_info.major >= 3:
                        ret.append(pkl.load(f, encoding='latin1'))
                    else:
                        ret.append(pkl.load(f))
            except EOFError:
                pass

        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    except pkl.UnpicklingError:
        with gzip.open(file_name, 'rb') as f:
            try:
                while True:
                    # Python 3 needs explicit encoding specification,
                    # which Python 2 lacks:
                    if sys.version_info.major >= 3:
                        ret.append(pkl.load(f, encoding='latin1'))
                    else:
                        ret.append(pkl.load(f))
            except EOFError:
                pass

        if len(ret) == 1:
            return ret[0]
        else:
            return ret


def h5Save(data, file_name, keys='None', compress=True):
    """
    Writes one or many arrays to h5 file

    data:
      single array to save or iterable of arrays to save.
      For iterable all arrays are written to the file.
    file_name: string
      path/name of destination file
    keys: string / list thereof
      For single arrays this is a single string which is used as a name for the data set.
      For multiple arrays each dataset is named by the corresponding key.
      If keys is ``None``, the dataset names created by enumeration: ``data%i``
    compress: Bool
      Whether to use lzf compression, defaults to ``True``. Most useful for label arrays.
    """

    compr = 'lzf' if compress else None
    f = h5py.File(file_name, "w")
    if isinstance(data, list) or isinstance(data, tuple):  #hasattr(keys, '__len__') and not isinstance(keys, str):
        for i, d in enumerate(data):
            f.create_dataset(keys[i], data=d, compression=compr)
    else:
        if keys is None:
            f.create_dataset('data', data=data, compression=compr)
        else:
            f.create_dataset(keys, data=data, compression=compr)
    f.close()


def h5Load(file_name, keys=None):
    """
    Loads data sets from h5 file

    file_name: string
      destination file
    keys: string / list thereof
      Load only data sets specified in keys and return as list in the order of ``keys``
      For a single key the data is returned directly - not as list
      If keys is ``None`` all datasets that are listed in the keys-attribute of the h5 file are loaded.
    """
    ret = []
    try:
        f = h5py.File(file_name, "r")
    except IOError:
        raise IOError("Could not open h5-File %s" % (file_name))

    if keys is not None:
        if isinstance(keys, str):
            ret.append(f[keys].value)
        else:
            for k in keys:
                ret.append(f[k].value)
    else:
        for k in f.keys():
            ret.append(f[k].value)

    f.close()

    if len(ret) == 1:
        return ret[0]
    else:
        return ret


def timeit(foo, n=1):
    """Decorator: decorates foo such that its execution time is printed upon call"""

    def decorated(*args, **kwargs):
        t0 = time.time()
        if n > 1:
            for i in xrange(n - 1):
                foo(*args, **kwargs)

        ret = foo(*args, **kwargs)
        t = time.time() - t0
        if hasattr(foo, '__name__'):
            print "Function <%s> took %.5f s averaged over %i execs" % (foo.__name__, t / n, n)
        else:
            print "Function took %.5f s averaged over %i execs" % (t / n, n)
        return ret

    return decorated


if __name__ == "__main__":
    #  c = np.array([1,2,3,4,5,6,7,8])
    #  _SMA(c,8)
    #  names = ['NewBase', 'NewDeep', 'NewBaseLR', 'NewDeepBlind']
    #  previewDiffPlot(names, block_name=3, number=23)
    #   plotInfoFromFiles("/home/mfk/CNN_Training/3D/NewDeepBlind/", "NewDeepBlind")
    #   plotInfoFromFiles("/home/mfk/CNN_Training/MLP/Piano", "Piano")
    plotInfoFromFiles("/home/mfk/CNN_Training/MLP/MNIST_MLP_warp", "MNIST_MLP_warp")
#  print "Testing CNN-Training Configuration and Backup of Scripts"
#  conf = ConfigObj("../Quick_config.py", 0, "../elektronn-train")

#  import scipy.ndimage as nd
#  g = np.random.rand(200,200)
#  g = nd.gaussian_filter(g, 1)
#  b = np.random.rand(200,200)
#  b = nd.gaussian_filter(b, 1)
#  r = np.minimum(g, b)
#  img = np.dstack([r,g,b])
#  plt.imshow(img, interpolation='none')



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import numpy as np


class _Layer(object):
    def __init__(self, input, filter=1, pool=1, stride=1, mfp=True):
        self.out = input - filter + 1
        if self.out <= 0:
            raise ValueError('CNN has no output for layer with input', input)
        self.stride = stride
        rest = self.out % pool
        self.pool_out = self.out // pool
        if pool > 1:
            if mfp and rest != 1:
                raise ValueError('MFP fails for layer with input', input)
            elif not mfp and rest > 0:
                raise ValueError('Uneven Pools for layer with input', input)

    def setfield(self, field):
        self.field = field
        self.overlap = field - self.stride


class _CNNCalculator(object):
    def __init__(self, filters, poolings, desired_input, MFP, force_center,
                 desired_output):
        self.mfp = MFP
        self.fields = self.getFields(filters, poolings)
        fow = self.fields[-1]
        if fow % 2 == 0:
            if force_center:
                raise ValueError('Receptive Fields are not centered with field of view (%i)' % fow)
            else:
                print 'WARNING: Receptive Fields are not centered with even field of view (%i)' % fow
        self.offset = float(fow) / 2

        valid_inputs = []
        valid_outputs = []
        for inp in xrange(2, 5000):
            try:
                self.calclayers(inp, filters, poolings, MFP)
                valid_inputs.append(inp)
                valid_outputs.append(self.out[-1])
            except:
                pass

        if desired_output is not None:
            if desired_output in valid_outputs:
                i = valid_outputs.index(desired_output)
                input = valid_inputs[i]
            else:
                valid_outputs = np.array(valid_outputs)
                input = valid_outputs[valid_outputs <= desired_output][-1]
                print "Info: output size requires input>1200, next smaller output (%i) is used" % input
                valid_outputs = list(valid_outputs)

            i = valid_outputs.index(desired_output)  # input corresponding to that output
            input = valid_inputs[i]

        elif desired_input in valid_inputs:
            input = desired_input
        elif desired_input is None:
            input = valid_inputs[-1]
        else:
            valid_inputs = np.array(valid_inputs)
            if desired_input < valid_inputs[0]:
                input = valid_inputs[0]
                print "Info: input (%i) changed to (%i) (size too small)" % (desired_input, input)
            else:
                input = valid_inputs[valid_inputs <= desired_input][-1]
                print "Info: input (%i) changed to (%i) (size not possible)" % (desired_input, input)
                valid_inputs = list(valid_inputs)

        self.valid_inputs = valid_inputs
        self.calclayers(input, filters, poolings, MFP)
        self.input = input
        self.pred_stride = self.layers[-1].stride
        for lay, field in zip(self.layers, self.fields):
            lay.setfield(field)

        self.overlap = [l.overlap for l in self.layers]

    def calclayers(self, input, filters, poolings, mfp):
        stride = poolings[0]
        self.layers = [_Layer(input, filters[0], poolings[0], stride, mfp=mfp[0])]
        for i in xrange(1, len(filters)):
            stride = np.multiply(stride, poolings[i])
            lay = _Layer(self.layers[i - 1].pool_out, filters[i], poolings[i], stride, mfp[i])
            self.layers.append(lay)

        self.pool_out = [l.pool_out for l in self.layers]
        self.out = [l.out for l in self.layers]
        self.stride = [l.stride for l in self.layers]

    def __repr__(self):
        ls = self.pool_out[::-1] if not isinstance(self.pool_out[0], list) else zip(*self.pool_out)[::-1]
        out = self.out[::-1] if not isinstance(self.out[0], list) else zip(*self.out)[::-1]
        fields = self.fields[::-1] if not isinstance(self.fields[0], list) else zip(*self.fields)[::-1]
        stride = self.stride[::-1] if not isinstance(self.stride[0], list) else zip(*self.stride)[::-1]
        overlap = self.overlap[::-1] if not isinstance(self.overlap[0], list) else zip(*self.overlap)[::-1]
        s = "Input: "+repr(self.input)+"\nLayer/Fragment sizes:\t"+repr(ls)+"\nUnpooled Layer sizes:\t"+repr(out)+\
        "\nReceptive fields:\t"+repr(fields)+"\nStrides:\t\t"+repr(stride)+\
        "\nOverlap:\t\t"+repr(overlap)+"\nOffset:\t\t"+repr(self.offset)+".\nIf offset is non-int: output neurons lie centered on input neurons,they have an odd FOV\n"
        return s

    def getFields(self, filter, pool):
        def recFields_helper(filter, pool):
            rf = [None] * (len(filter) + 1)
            rf[-1] = 1
            for i in xrange(len(filter), 0, -1):
                rf[i - 1] = rf[i] * pool[i - 1] + filter[i - 1] - 1
            return rf[0]

        fields = []
        for i in xrange(1, len(filter) + 1):
            fields.append(recFields_helper(filter[:i], pool[:i]))

        return fields


class _multiCNNCalculator(_CNNCalculator):
    """ Adaptor Class to unify multiple CNNCalculators"""

    def __init__(self, calcs):
        self.fields = []
        self.offset = []
        self.valid_inputs = []
        self.input = []
        self.pred_stride = []
        self.stride = []
        self.pool_out = []
        self.out = []
        self.overlap = []
        for c in calcs:
            self.fields.append(c.fields)
            self.offset.append(c.offset)
            self.valid_inputs.append(c.valid_inputs)
            self.input.append(c.input)
            self.pred_stride.append(c.pred_stride)
            self.overlap.append(c.overlap)
            self.pool_out.append(c.pool_out)
            self.out.append(c.out)
            self.stride.append(c.stride)


def CNNCalculator(filters,
                  poolings,
                  desired_input=None,
                  MFP=False,
                  force_center=False,
                  desired_output=None,
                  n_dim=1):
    """
    Helper to calculate CNN architectures

    This is a *function*, but it returns an *object* that has various architecture values as attributes.
    Useful is also to simply print 'd' as in the example.

    Parameters
    ----------

    filters: list
      Filter shapes (for anisotropic filters the shapes are again a list)
    poolings: list
      Pooling factors
    desired_input: int or list of int
      Desired input size(s). If ``None`` a range of suggestions can be found in the attribute ``valid_inputs``
    MFP: list of int/{0,1}
      Whether to apply Max-Fragment-Pooling in this layer and check compliance with max-fragment-pooling
      (requires other input sizes than normal pooling)
    force_center: Bool
      Check if output neurons/pixel lie at center of input neurons/pixel (and not in between)
    desired_output: int or list of int
      Alternative to ``desired_input``
    n_dim: int
      Dimensionality of CNN

    Examples
    --------

    Calculation for anisotropic "flat" 3d CNN with MFP in the first layers only::

      >>> desired_input   = [211, 211, 20]
      >>> filters         = [[6,6,1], [4,4,4], [2,2,2], [1,1,1]]
      >>> pool            = [[2,2,1], [2,2,2], [2,2,2], [1,1,1]]
      >>> MFP             = [1,        1,       0,       0,   ]
      >>> n_dim=3
      >>> d = CNNCalculator(filters, pool, desired_input, MFP=MFP, force_center=True, desired_output=None, n_dim=n_dim)
      Info: input (211) changed to (210) (size not possible)
      Info: input (211) changed to (210) (size not possible)
      Info: input (20) changed to (22) (size too small)
      >>> print d
      Input: [210, 210, 22]
      Layer/Fragment sizes:	[[102, 49, 24, 24], [102, 49, 24, 24], [22, 9, 4, 4]]
      Unpooled Layer sizes:	[[205, 99, 48, 24], [205, 99, 48, 24], [22, 19, 8, 4]]
      Receptive fields:	[[7, 15, 23, 23], [7, 15, 23, 23], [1, 5, 9, 9]]
      Strides:		[[2, 4, 8, 8], [2, 4, 8, 8], [1, 2, 4, 4]]
      Overlap:		[[5, 11, 15, 15], [5, 11, 15, 15], [0, 3, 5, 5]]
      Offset:		[11.5, 11.5, 4.5].
          If offset is non-int: floor(offset).
          Select labels from within img[offset-x:offset+x]
          (non-int means, output neurons lie centered on input neurons,
          i.e. they have an odd field of view)
    """

    assert len(poolings) == len(filters)

    if MFP is False:
        MFP = [False, ] * len(filters)

    if n_dim == 1:  #not hasattr(filters[0], '__len__') :
        return _CNNCalculator(filters, poolings, desired_input, MFP, force_center, desired_output)
    else:
        if desired_input is None:
            desired_input = (None, ) * n_dim
        elif not hasattr(desired_input, '__len__'):
            desired_input = (desired_input, ) * n_dim
        if desired_output is None:
            desired_output = (None, ) * n_dim
        elif not hasattr(desired_output, '__len__'):
            desired_output = (desired_output, ) * n_dim
        if not hasattr(poolings[0], '__len__'):
            poolings = [[p, ] * n_dim for p in poolings]
        if not hasattr(filters[0], '__len__'):
            filters = [[f, ] * n_dim for f in filters]
        if not hasattr(MFP[0], '__len__'):
            MFP = [[m, ] * n_dim for m in MFP]

        assert len(MFP) == len(filters)

        filters = [list(l) for l in zip(*filters)]
        poolings = [list(l) for l in zip(*poolings)]
        MFP = [list(l) for l in zip(*MFP)]

        calcs = []
        for f, p, d, o, mfp in zip(filters, poolings, desired_input, desired_output, MFP):
            c = _CNNCalculator(f, p, d, mfp, force_center, o)
            calcs.append(c)

        return _multiCNNCalculator(calcs)


def initWeights(shape, scale='glorot', mode='normal', pool=None):
    if len(shape) == 1:
        n_in = shape[0]
    if len(shape) == 2:
        n_in, n_out = shape[0], shape[1]
    elif len(shape) == 4:
        n_in = np.float(np.prod(shape[1:]))
        n_out = np.float((shape[0] * np.prod(shape[2:]) / np.prod(pool)))
    elif len(shape) == 5:
        n_in = np.float(np.prod(shape[1:]))
        n_out = np.prod(shape[0:2]) * np.prod(shape[3:]) / np.prod(pool)

    if mode == 'const':
        W = np.ones(shape) * scale
    elif mode == 'rnn':
        assert (shape[0] == shape[1])
        W = np.random.uniform(-1.0, 1.0, size=shape)
        U, s, V = np.linalg.svd(W)
        #W = U
        W = W / s[0]  ###TODO which is better?
    elif mode == 'fix-uni':
        W = np.random.uniform(-scale, scale, shape)
    elif scale == 'glorot':
        W_scale = np.sqrt(2.0 / (n_in + n_out))
        if mode == 'normal':
            W = np.random.normal(0, W_scale, shape)
        elif mode == 'uni':
            W = np.random.uniform(-W_scale, W_scale, shape)
    else:
        raise ValueError("Invalid weigh initialisation parameters")
    return W


if __name__ == "__main__":
    print "Testing CNNCalculator"
    desired_input = [
        180, 180, 30
    ]  # (*) <int> or <2/3-tuple> in (x,y)/(x,y,z)-order for anisotropic CNN
    filters = [[6, 6, 1],
               [4, 4, 4],
               [4, 4, 4],
               [4, 4, 4],
               ]  #[4,4,1], [2,2,4], [2,2,4], [2,2,2], [1,1,1]] # [1,1,1]
    pool = [[2, 2, 1],
            [2, 2, 2],
            [2, 2, 1],
            [1, 1, 1],
            ]  # [1,1,1], [1,1,1], [1,1,2], [1,1,1], [1,1,1]] # [1,1,1]
    MFP = [True, ] * 3 + [False, ] * 1

    n_dim = 3
    d = CNNCalculator(filters,
                      pool,
                      desired_input,
                      MFP=MFP,
                      force_center=False,
                      desired_output=None,
                      n_dim=n_dim)
    print d



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import numpy as np
import theano
from theano import tensor as T


def maxabs(t1, t2):
    pos = T.where(t1 > t2, t1, t2)
    neg = T.where(-t1 > -t2, t1, t2)
    ret = T.where(pos >= -neg, pos, neg)
    return ret


def my_max_pool_3d(sym_input, pool_shape=(2, 2, 2)):
    """ this one is pure theano. Hence all gradient-related stuff is working! No dimshuffling"""

    s = None
    if pool_shape[2] > 1:
        for i in xrange(pool_shape[2]):
            t = sym_input[:, :, :, :, i::pool_shape[2]]
            if s is None:
                s = t
            else:
                s = T.maximum(s, t)
    else:
        s = sym_input

    if pool_shape[0] > 1:
        temp = s
        s = None
        for i in xrange(pool_shape[0]):
            t = temp[:, i::pool_shape[0], :, :, :]
            if s is None:
                s = t
            else:
                s = T.maximum(s, t)

    if pool_shape[1] > 1:
        temp = s
        s = None
        for i in xrange(pool_shape[1]):
            t = temp[:, :, :, i::pool_shape[1], :]
            if s is None:
                s = t
            else:
                s = T.maximum(s, t)
    sym_ret = s

    return sym_ret


def maxout(conv_out, factor=2, mode='max', axis=1):
    """
    Pools axis 1 (the channels) of ``conv_out`` by ``factor``.
    I.e. the number of channels is decreased by this factor.
    The pooling can either be done as ``max`` or ``maxabs``.
    Spatial dimensions are unchanged
    """
    assert factor > 1
    if mode == 'max':
        compare = T.maximum
    elif mode == 'maxabs':
        compare = maxabs

    if axis == 1:
        ret = conv_out[:, 0::factor]
        for i in xrange(1, factor):
            t = conv_out[:, i::factor]
            ret = compare(ret, t)

    elif axis == 2:
        ret = conv_out[:, :, 0::factor]
        for i in xrange(1, factor):
            t = conv_out[:, :, i::factor]
            ret = compare(ret, t)

    return ret


def pooling2d(conv_out, pool_shape=(2, 2), mode='max'):
    """
    Pools axis 2,3 (x,y) of ``conv_out`` by respective ``pool_shape``.
    I.e. the spatial extent is decreased by this factor.
    The pooling can either be done as ``max`` or ``maxabs``.
    """
    if mode == 'max':
        compare = T.maximum
    elif mode == 'maxabs':
        compare = maxabs

    ret = conv_out

    if pool_shape[1] > 1:
        ret = conv_out[:, :, :, 0::pool_shape[1]]
        for i in xrange(1, pool_shape[1]):
            t = conv_out[:, :, :, i::pool_shape[1]]
            ret = compare(ret, t)

    if pool_shape[0] > 1:
        accum = ret[:, :, 0::pool_shape[0], :]
        for i in xrange(1, pool_shape[0]):
            t = ret[:, :, i::pool_shape[0], :, ]
            accum = compare(accum, t)

        ret = accum

    return ret


def pooling3d(conv_out, pool_shape=(2, 2, 2), mode='max'):
    """
    Pools axis 2,3 (x,y) of ``conv_out`` by respective ``pool_shape``.
    I.e. the spatial extent is decreased by this factor.
    The pooling can either be done as ``max`` or ``maxabs``.
    """
    if mode == 'max':
        compare = T.maximum
    elif mode == 'maxabs':
        compare = maxabs

    ret = conv_out  #(1,4,1,4,4)

    if pool_shape[2] > 1:
        ret = conv_out[:, :, :, :, 0::pool_shape[2]]
        for i in xrange(1, pool_shape[2]):
            t = conv_out[:, :, :, :, i::pool_shape[2]]
            ret = compare(ret, t)

    if pool_shape[0] > 1:
        accum = ret[:, 0::pool_shape[0], :, :, :]
        for i in xrange(1, pool_shape[0]):
            t = ret[:, i::pool_shape[0], :, :, :]
            accum = compare(accum, t)

        ret = accum

    if pool_shape[1] > 1:
        accum = ret[:, :, :, 0::pool_shape[1], :]
        for i in xrange(1, pool_shape[1]):
            t = ret[:, :, :, i::pool_shape[1], :]
            accum = compare(accum, t)

        ret = accum

    return ret


if __name__ == "__main__":
    import sys, os
    sys.path.append(os.path.expanduser("~/devel/ELEKTRONN/"))
    from elektronn.training.trainutils import timeit
    sym_input = T.TensorType(dtype=theano.config.floatX, broadcastable=[False] * 5)()
    pool = (2, 2, 2)

    sym_ret = my_max_pool_3d(sym_input, pool)  #my_max_pool_3d_stupid(sym_input)
    f_maxp_3d_ref = timeit(theano.function([sym_input], sym_ret), 4)

    sym_ret = pooling3d(sym_input, pool, 'max')  #my_max_pool_3d_stupid(sym_input)
    f_maxp_3d = timeit(theano.function([sym_input], sym_ret), 4)

    sym_ret = pooling3d(sym_input, pool, 'maxabs')  #my_max_pool_3d_stupid(sym_input)
    f_maxp_3d_new = timeit(theano.function([sym_input], sym_ret), 4)

    #    sym_ret = maxabsPool3d(sym_input, pool)#my_max_pool_3d_stupid(sym_input)
    #    f_maxabsp_3d= timeit(theano.function([sym_input],sym_ret), 4)

    for i in range(0, 10, 2):
        inp = np.random.rand(1, 16 + 32 * i, 1, 16 + 32 * i, 16 + 32 *
                             i).astype('float32') - 0.5
        print inp.shape
        a0 = f_maxp_3d_ref(inp)
        a1 = f_maxp_3d(inp)
        a2 = f_maxp_3d_new(inp)
        #a3 = f_maxabsp_3d(inp)
        assert np.all(np.equal(a1, a0))
        #print a1.shape

    r1 = a1[0, :, 0]
    r2 = a2[0, :, 0]
    #r3 = a3[0,:,0]
    r0 = inp[0, :, 0]
    #d  = r3 - r4



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban


def pprinttime(t):
    """Custom printing of elapsed time"""
    if t > 4000:
        s = 't=%.1fh' % (t / 3600)
    elif t > 300:
        s = 't=%.0fm' % (t / 60)
    else:
        s = 't=%.0fs' % (t)
    return s



# -*- coding: utf-8 -*-
"""
ELEKTRONN - Neural Network Toolkit
Copyright (c) 2015 Gregor Urban, Marius Killinger
"""

from elektronn.training.traindata import MNISTData
from elektronn.net.convnet import MixedConvNN

# Load Data #
data = MNISTData(path=None,
                 convert2image=True,
                 shift_augment=False)

# Create CNN #
batch_size = 100
cnn = MixedConvNN((28, 28), input_depth=1)
cnn.addConvLayer(10,
                 5,
                 pool_shape=2,
                 activation_func="abs")  # (nof, filtersize)
cnn.addConvLayer(8, 5, pool_shape=2, activation_func="abs")
cnn.addPerceptronLayer(100, activation_func="abs")
cnn.addPerceptronLayer(80, activation_func="abs")
cnn.addPerceptronLayer(
    10,
    activation_func="abs"
)  # need 10 outputs as there are 10 classes in the data set
cnn.compileOutputFunctions()
cnn.setOptimizerParams(SGD={'LR': 1e-2, 'momentum': 0.9}, weight_decay=0)

print "training..."
for i in range(5000):
    d, l = data.getbatch(batch_size)
    loss, loss_instance, time_per_step = cnn.trainingStep(d, l, mode="SGD")

    if i % 100 == 0:
        valid_loss, valid_error, valid_predictions = cnn.get_error(data.valid_d, data.valid_l)
        print "update:", i, "; Validation loss:", valid_loss, "Validation error:", valid_error * 100., "%"

loss, error, test_predictions = cnn.get_error(data.test_d, data.test_l)
print "Test loss:", loss, "Test error:", error * 100., "%"
print "Done."



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban
"""
Supplementary functions to plot various CNN states
"""

import numpy as np
import matplotlib.pyplot as plt


def plotFilters(cnn, layer=0, channel=None, normalize=False, savename='filters_layer0.png'):
    W = cnn.layers[layer].W.get_value()

    if W.shape[1] == 3:  # RGB
        r = embedMatricesInGray(W[:, 0, :, :], normalize=normalize)
        g = embedMatricesInGray(W[:, 1, :, :], normalize=normalize)
        b = embedMatricesInGray(W[:, 2, :, :], normalize=normalize)
        mat = np.dstack((r, g, b))
        lower = min(0, mat.min())
        spread = mat.max() - lower
        mat = mat - lower
        mat = 255.0 / spread * mat
        mat = mat.astype('uint8')

    elif W.shape[1] == 1 or channel is None:  # gray-scale
        W = W[:, 0, :, :]
        mat = embedMatricesInGray(W, normalize=normalize)

    elif W.shape[1] == 1 or channel is not None:  # gray-scale
        W = W[:, channel, :, :]
        mat = embedMatricesInGray(W, normalize=normalize)

    plt.figure()
    plt.imshow(mat, interpolation='None', cmap='gray')  # Note that cmap is ignored by matplotlib for RGB-img
    plt.show()
    print "Saving filter image as as %s" % (savename)
    plt.savefig(savename, bbox_inches='tight')


def showActivations(cnn, data, show_first_class_prob=False, no_show=False):
    """
    Plots activation maps given data. It requires that cnn.debug_functions contains a list of functions
    that return the activations (i.e. cnn.compileDebugFunctions must have been called

    Parameters
    ----------
    
    cnn: 
      instance of MixedConvNN
    data:
      input to cnn for which activations should be shown
    show_first_class_prob:
      True/False whether to additionally show the probability map for the first class
    no_show:
      True/False whether to pop up plots or silently return a list of image arrays
    
    """
    n = len(cnn.layers)
    all_act = []
    myfig = plt.figure(figsize=(15, 14))

    if len(cnn.debug_functions) == 0:
        print "No debug functions found in cnn! Compile them first"

    for layer, dbgf, i in zip(cnn.layers, cnn.debug_functions, range(n)):
        activity = dbgf(data)[
            0
        ]  # only first image in batch (nof, x-dim, y-dim) or (z-dim, nof, x-dim, y-dim)
        sp = activity.shape

        if len(sp) != 3:
            if len(sp) == 4:
                activity = activity.reshape(sp[0] * sp[1], sp[2], sp[3])  #np.swapaxes(np.swapaxes(activity,2,0),2,4)
            else:
                continue
        if len(sp) == 3:
            image = embedMatricesInGray(activity, 1, True)
        elif len(sp) == 5:
            image = embedMatricesInGray(activity, 1, True, fixed_n_horizontal=sp[1])

        all_act.append(image)
        showMultipleFiguresAdd(myfig, n, i, image, "Layer-" + str(i) + " Activity")

    plt.tight_layout()

    if show_first_class_prob:
        prob_pred = cnn.class_probabilities(data)
        plt.figure()
        plt.gray()
        plt.title('prediction map of first class')
        plt.imshow(prob_pred[:, 0].reshape(np.sqrt(np.shape(prob_pred)[0]), -1), interpolation='nearest')

    if not no_show:
        plt.show()

    return all_act


def showParamHistogram(cnn, no_show=False, onlyW=True):
    """
    Plots histograms of parameter/weight values.

    Parameters
    ----------

    :type no_show: object
    cnn:
      instance of MixedConvNN
    onlyW:
      True/False whether to ignore the biases b
    no_show:
      True/False whether to pop up plots or silently return a list of image arrays

    """
    n = len(cnn.layers)
    for layer, i in zip(cnn.layers, range(n)):
        for par in layer.params if onlyW == False else [layer.W]:
            pa = par.get_value()
            print 'Layer {0:2d}: mean={1:3.6e}, median={2:3.6e}, mean(abs)={3:2.6e}, median(abs)={4:2.6e}'.format(
                1 + i, np.mean(pa), np.median(pa), np.mean(abs(pa)), np.median(abs(pa)))
            print 'Layer  ' + str(i + 1) + ': filter_shape =', np.shape(pa)

            hist, bins = np.histogram(pa.flatten(), bins=100)
            width = 0.7 * (bins[1] - bins[0])
            center = (bins[:-1] + bins[1:]) / 2
            plt.figure()
            plt.bar(center, hist, align='center', width=width)
            plt.xlabel(str(np.shape(pa)) + ' parameters ', fontsize=16)
            plt.ylabel('# (Layer ' + str(i + 1) + ')' + str(), fontsize=16)

        if not no_show:
            plt.show()


def showActivityHistogram(cnn, data, no_show=False):
    """
    Plots histograms of activation maps given data. It requires that cnn.debug_functions contains a list
    of functions that return the activations (i.e. cnn.compileDebugFunctions must have been called

    Parameters
    ----------

    cnn:
      instance of MixedConvNN
    data:
      input to cnn for which activations should be shown
    no_show:
      True/False whether to pop up plots or silently return a list of image arrays

    """
    n = len(cnn.layers)
    for layer, dbgf, i in zip(cnn.layers, cnn.debug_functions, range(n)):
        activity = dbgf(data).flatten()
        hist, bins = np.histogram(activity, bins=100)
        width = 0.7 * (bins[1] - bins[0])
        center = (bins[:-1] + bins[1:]) / 2
        plt.figure()
        plt.bar(center, hist, align='center', width=width)

        plt.xlabel(str(len(activity)) + ' activations ', fontsize=16)
        plt.ylabel('# (Layer ' + str(i + 1) + ')' + str(), fontsize=16)
    if not no_show:
        plt.show()


def embedMatricesInGray(mat, border_width=1, normalize=False, output_ratio=1.7, fixed_n_horizontal=0):
    """
    Creates a big matrix out of smaller ones (mat) assumed format of mat:(index, ch, i_vert,i_horiz)
    """

    #    sh = np.shape(mat)
    #    if sh[1]==3: # interpret 3 channels as rgb
    #      mat = np.swapaxes(mat, 1, 3) # swap RGB channels at back
    #    if sh[1]==1: # work directly on this channel
    #      mat = mat[:,0]
    #    else: # create an image for each channel
    #      channels = []
    #      for i in xrange(sh[1]):
    #        ch = embedMatricesInGray(mat[:,i], border_width, normalize, output_ratio, fixed_n_horizontal)
    #        channels.append(ch)
    #        return channels
    sh = np.shape(mat)
    assert len(sh) == 3
    n = sh[0]
    if fixed_n_horizontal > 0:
        nhor = fixed_n_horizontal
    else:
        nhor = int(np.sqrt(n * output_ratio))  # aim: ratio 16:9
    nvert = int(n * 1.0 / nhor + 1)  #warning: too big: nvert*nhor >= n

    ret = np.zeros((nvert * (border_width + sh[1]), nhor * (border_width + sh[2])), dtype=np.float32)

    if normalize:
        maxs = [np.max(mat[i, :, :]) + 1e-8 for i in range(n)]
        mins = [np.min(mat[i, :, :]) for i in range(n)]
    else:
        maxs = [1] * n
        mins = [0] * n

    for j in range(nvert):
        for i in range(nhor):
            if i + j * nhor >= n:
                return ret
            #print (j*(border_width+sh[1]),j*(border_width+sh[1])+sh[1], i*(border_width+sh[2]),i*(border_width+sh[2])+sh[2])
            #print shape(mat[i+j*nhor,:,:])
            ret[j * (border_width + sh[1]):j * (border_width + sh[1]) + sh[1],
                i * (border_width + sh[2]):i * (border_width + sh[2]) + sh[2]
               ] = (mat[i + j * nhor, :, :] - mins[i + j * nhor]) / (maxs[i + j * nhor] - mins[i + j * nhor])
    return ret


def showMultipleFiguresAdd(fig, n, i, image, title, isGray=True):
    """Add <i>th (of n, start: 0) image to figure <fig> as subplot (GRAY)"""

    x = int(np.sqrt(n) + 0.9999)
    y = int(n / x + 0.9999)
    if (x * y < n):
        if x < y:
            x += 1
        else:
            y += 1

    ax = fig.add_subplot(x, y, i)  #ith subplot in grid x,y
    ax.set_title(title)
    if isGray:
        plt.gray()
    ax.imshow(image, interpolation='nearest')



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import os
from elektronn.training.config import default_config, Config  # the global user-set config

# prevent Qt-backend on remote machines early! (other modules may import mpl)
#no_X = default_config.no_X
#hostname = socket.gethostname()    
#if hostname in default_config.no_X_hosts or no_X:
#  print "Importing Matplotlib without interactive backend, plots can only be saved to files in this session!"
#  import matplotlib
#  matplotlib.use('AGG')


def create_predncnn(config_file,
                    n_ch,
                    n_lab,
                    gpu=None,
                    override_MFP_to_active=False,
                    imposed_input_size=None,
                    param_file=None):
    """
    Creates and compiles a CNN/NN as specified in a config file (used for training).
    Loads the last parameters from the training directory.

    The CNN/NN object is returned

    Parameters
    ----------

    config_file: string
      Path to a CNN config file
    n_ch: int
      Number of input channels, for a MLP this is the dimensionality of the input vectors, for CNNs this is the
      number of channels in an image/volume (e.g. 1 for plain gray value images)
    n_lab: int
      Number of distinct labels/classes
    gpu: int
      Specifying id of GPU to initialise for usage. E.g. 1 --> "gpu1", None will initialise gpu0,\
      False will not initialise any GPU. This only works if "device" is not set in ``.theanorc`` or if theano
      has not been imported up to now. If the initialisation fails an error will be printed but the script
      will not crash.
    override_MFP_to_active: Bool
      If true, activates MFP in all layers where possible, ignoring the configuration in the config file.
      This is useful for prediction using a config file from training. (only for CNN)
    imposed_input_size: tuple or None
      Similar as above, this can be used to impose another input size than specified in the config file.
      (z,x,y)!!! (only for CNN)
    param_file: string/None
      If other parameters than "*-Last.param" should be loaded, this can specify the param file.
    """
    config_file = os.path.expanduser(config_file)
    if gpu == None:
        gpu = default_config.device

    config = Config(config_file,
                    gpu,
                    None,
                    use_existing_dir=True,
                    override_MFP_to_active=override_MFP_to_active,
                    imposed_input_size=imposed_input_size)  # inits gpu

    from elektronn.net.netcreation import createNet  # import after gpu init

    os.chdir(config.save_path)  # The trainer works directly in the save dir
    cnn = createNet(config, config.dimensions.input, n_ch, n_lab, config.dimensions)  # 1 ch 2 label
    if param_file is None:
        path = "%s-LAST.param" % config.save_name
    else:
        path = os.path.expanduser(param_file)

    cnn.loadParameters(path, strict=True)

    return cnn


if __name__ == "__main__":
    import numpy as np
    path = "~/devel/ELEKTRONN/Other/config_files/I3-Deeper-thin.py"
    cnn = create_predncnn(path,
                          1,
                          2,
                          gpu=0,
                          override_MFP_to_active=True,
                          imposed_input_size=(60, 300, 300))
    x = np.random.rand(1, 800, 800, 200).astype(np.float32)
    y = cnn.predictDense(x)



# -*- coding: utf-8 -*-
"""
ELEKTRONN - Neural Network Toolkit

Copyright (c) 2014 - now
Max-Planck-Institute for Medical Research, Heidelberg, Germany
Authors: Marius Killinger, Gregor Urban
"""

import matplotlib.pyplot as plt

from elektronn.training.traindata import MNISTData
from elektronn.net.convnet import MixedConvNN
from elektronn.net.introspection import embedMatricesInGray

# Load Data #
data = MNISTData(path=None,
                 convert2image=False,
                 shift_augment=False)

# Create Autoencoder #
batch_size = 100
cnn = MixedConvNN((28**2), input_depth=None)
cnn.addPerceptronLayer(n_outputs=300, activation_func="tanh")
cnn.addPerceptronLayer(n_outputs=200, activation_func="tanh")
cnn.addPerceptronLayer(n_outputs=50, activation_func="tanh")
cnn.addTiedAutoencoderChain(n_layers=None,
                            activation_func="tanh",
                            input_noise=0.3)
cnn.compileOutputFunctions(target="regression")  #compiles the cnn.get_error function as well
cnn.setOptimizerParams(SGD={'LR': 5e-1, 'momentum': 0.9}, weight_decay=0)

print "training..."
for i in range(40000):
    d, l = data.getbatch(batch_size)
    loss, loss_instance, time_per_step = cnn.trainingStep(d, d, mode="SGD")

    if i % 100 == 0:
        print "update:", i, "; Training error:", loss

cnn.layers[3].input_noise.set_value(0.0)
loss, _, test_predictions = cnn.get_error(data.valid_d.reshape((10000, 784)), data.valid_d.reshape((10000, 784)))
print "Final error:", loss
print "Done."

plt.figure(figsize=(14, 6))
plt.subplot(121)
images = embedMatricesInGray(data.valid_d[:200].reshape((200, 28, 28)), 1)
plt.imshow(images, interpolation='none', cmap='gray')
plt.title('Data')

plt.subplot(122)
recon = embedMatricesInGray(test_predictions[:200].reshape((200, 28, 28)), 1)
plt.imshow(recon, interpolation='none', cmap='gray')
plt.title('Reconstruction')

d_v = data.valid_d
pred = test_predictions.clip(d_v.min(), d_v.max())
recon = embedMatricesInGray(pred[:200].reshape((200, 28, 28)), 1)
plt.imshow(recon, interpolation='none', cmap='gray')

# Use pre-trained weights from AE to train a classifier
cnn.saveParameters('AE-pretraining.param', layers=cnn.layers[0:3])  # only save the weights from the compression

cnn2 = MixedConvNN((28**2), input_depth=None)
cnn2.addPerceptronLayer(n_outputs=300, activation_func="tanh")
cnn2.addPerceptronLayer(n_outputs=200, activation_func="tanh")
cnn2.addPerceptronLayer(n_outputs=50, activation_func="tanh")
cnn2.addPerceptronLayer(n_outputs=10, activation_func="tanh")
cnn2.compileOutputFunctions(target="nll"
                            )  #compiles the cnn.get_error function as well
cnn2.setOptimizerParams(SGD={'LR': 0.005, 'momentum': 0.9}, weight_decay=0)
cnn2.loadParameters('AE-pretraining.param')
print "training..."
for i in range(10000):
    d, l = data.getbatch(batch_size)
    loss, loss_instance, time_per_step = cnn2.trainingStep(d, l, mode="SGD")
    error = cnn2.get_error(d, l)[1]

    if i % 100 == 0:
        print "update:", i, "; Training error:", error, "NLL:", loss

print "Final error:", loss
print "Done."



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

# CONFIGURATION TEMPLATE FOR CNN/NN-TRAINING

### Pipeline Setup ### ---------------------------------------------------------

# (*) <String>: where to create the CNN directory. In this directory a new
# folder is created with the save name of the model
save_path = "~/CNN_Training/3D/"
# <Bool>: whether to create plots of the training progress
plot_on = True
# <Bool>: whether to print Training status to the console
print_status = True
# False/int: False --> use .theanorc value, int --> use gpu<i>
device = False
# <float> hours: frequency to save a permanent parameter snapshot
param_save_h = 1.0
# <float> hours: time after which first preview is made
initial_prev_h = 1.0
# <float> hours: frequency to create previews
prev_save_h = 3.0

### Paths and General ### ------------------------------------------------------

# (*) <String>: with the save_name.
save_name = 'Debug'
# <Bool>: whether to delete/overwrite existing directory
overwrite = True
save_path += save_name + '/'
# <String>/<None>: optional path to parameter file to initialise weights from
param_file = None

### Network Architecture ### ---------------------------------------------------
# Note: the last layer is added automatically (with n_lab outputs)

# <String> or <List> of <String>s: [relu], abs, linear, sig, tanh, (globally or per layer)
activation_func = 'relu'
# (*) <Int>: number of different slices/examples used for one optimisation step
batch_size = 1
# (*) <List> of <Float>>(0,1) or <Float>(0,1): "fail"-rates
# (globally or per layer). The last layer never has dropout (automatically).
dropout_rates = []

# Conv layers
# (*) 2/3:  for non image data (mode 'vect-scalar') this is ignored
n_dim = 2
# (*) <Int> or 2/3-Tuple: in (x,y)/(x,y,z)-order
desired_input = 100
# (*) <List> of <Int> or <List> of <2/3-tuples> or []: filter shapes in (x,y)/(x,y,z)-order
filters = []
# (*) <List> of <Int> or <List> of <2/3-tuples> or []: pool shapes in (x,y)/(x,y,z)-order
pool = []
# (*) <List> of <Int> / []: number of feature maps per layer
nof_filters = []
# ($) <String> or <List> of <String>: select pooling function
# (globally or per layer). Available: 'max', 'maxabs'
pooling_mode = 'max'
# (*) <List> of <Int>{0,1}/False: whether to apply Max-Fragment-Pooling (globally or per layer)
MFP = []

# Other
# ($) <Dict>/<None>: if in use: dict(n_hid=Int,  activation_func='tanh', iterations=None/Int)
rnn_layer_kwargs = None
# (*) <List> of <Int>: numbers of filters for perceptron layers (after conv layers)
MLP_layers = []
# <String>: 'nll' or 'regression'
target = 'nll'

### Data General ### -----------------------------------------------------------

# (*) <String>: combination of data and label types: 'img-img', 'img-scalar', vect-scalar'
mode = 'img-img'
# (*) <Int>/<None>: (None means auto detect, very slow, don't do this!)
n_lab = None
# <Bool>/<Int>: whether to "pre-fetch" batches in separate background
# process, <Bool> or number of processes (True-->2)
background_processes = False

### Data Images/CNN ### (the whole block is ignored for mode 'vect-scalar') ###
# (*) <String>: Path to data dir
data_path = '/'
# (*) <String>: Path to label dir
label_path = '/'
# (*) <List> of tuples: (file name, key of h5 data set)
d_files = []
# (*) <List> of tuples: (file name, key of h5 data set)
l_files = []
# <List>/<None>: sampling priorities for cubes or None, then:
# sampling ~ example_size. Will be normalised internally
cube_prios = None
# <List>: of cube indices (from the file-lists) to use as validation data, may be empty
valid_cubes = []
# <Float>: If the fraction of negative labels in an example patch exceeds
# this threshold this example is discarded
example_ignore_threshold = 0.0
# (*) <List>:  of integer channel-indices to apply grey augmentation, use []
# to disable. It distorts the histogram of the raw images
# (darker, lighter, more/less contrast)
grey_augment_channels = [0]
# ($) <Bool>: Whether to use weights for the examples (e.g. for Boosting-like training)
use_example_weights = False
# <Bool>: whether to flip/rotate/mirror data randomly (augmentation)
flip_data = True
# (*) <Bool>: if True 2D slices are only cut in z-direction, otherwise all 3 alignments are used
anisotropic_data = True
# ($) <Bool>: Special Training with lazy annotations
lazy_labels = False
# <Bool>/<Float>(0,1): Warping augmentations (CPU-intense, use background
# processes!). If <Float>: warping is applied to this fraction of
# examples e.g. 0.5 --> every 2nd example
warp_on = False
# <String>: Only applicable for *img-scalar*. If the CNN does not allow the
# original size of the images the following options are available:
# "crop"  : if img too big, cut the images to the next smaller valid input size,
# "keep"  : if img too big, keep the size and cut smaller patches with translations
# "0-pad" : if img to small, padd to the next bigger valid input with zeros,
# "c-pad" pad to the next bigger input with the average value of the border,
# "mirror": if img to small, padd to the next bigger valid input by mirroring along border
# "reject": if img size too small/big, throw exception
border_mode = "crop"

#  "standardise"/None: (0-mean, 1-std) (over all pixels)
pre_process = None
# <Bool>: set to True if data is in (z, (ch,) x, y) order, otherwise (ch, x, y, z) is assumed
zchxy_order = False
# <Bool>: If true, mirroring is only applied horizontally (e.g. for outdoor images or handwriting)
upright_x = False
# <Bool>/<Int> downsample trainig data by this factor (or not at all if False)
downsample_xy = False

### Data Preview ### (only for img-img) ----------------------------------------
# <String>/<None>: path to a h5-file that contains data to make preview
# predictions it must contain a list of image cubes (normalised between 0
# and 255) in the shape ((ch,) x,y,z)
preview_data_path = None
# <Dict>: specification of preview to create
preview_kwargs = dict(export_class=1, max_z_pred=5)

### Data Alternative / vect-scalar ### (this may replace the above CNN block) ###
# <String>: Name of Data Class in TrainData or <tuple>: (path_to_file, class_name)
data_class_name = None
# <Dict>: Arguments to init Data Class
data_load_kwargs = dict()
# <Dict>: Arguments for getbach method of Data Class (for training set only!).
#  The batch_size argument is added internally and needn't be specified here
data_batch_kwargs = dict()

### Optimisation Options ### ---------------------------------------------------
# (*) <Int>: number of update steps
n_steps = 10**12
# (*) <Int>: maximal Training time in seconds (overrides n_steps)
max_runtime = 24 * 3600
# (*) <List> of single <Int>: create plots, print status and test model after x steps
history_freq = [500]
# (*) <Int>: number of patches to test model on (valid and train subset)
monitor_batch_size = 10
# ($) False/<Float>: weighting of L2-norm on weights ("lambda"), False is equal to 0.0
weight_decay = False
# ($) <None> or <List> of <Float>: weights for the classes, will be normalised internally
class_weights = None
# ($) <None>/<Float>: if <Float> value is set, unlabelled examples (-1) will
# be trained on the current prediction if the predicted probability exceeds this threshold.
label_prop_thresh = None

# ($) <String>: [SGD]/CG/RPORP/LBFGS method for training
optimizer = 'SGD'
# (*) <Float>: decay of SGD learning rate w.r.t to an interval of 1000 update steps
LR_decay = 0.993
# (*) <List> of tuples/<None>: (#iteration, new_LR), if used this sets the LR at
# the specified iteration steps to the specified value. This is independent of the decay.
LR_schedule = None

# ------------------------------------------------------------------------------
### SGD
# (*) <Dict>: initial learning rate and momentum
SGD_params = dict(LR=0.001, momentum=0.9)

### RPROP ($)
RPROP_params = dict(penalty=0.35,
                    gain=0.2,
                    beta=0.7,
                    initial_update_size=1e-4)  # <Dict>

# CG ($)
CG_params       = dict(n_steps=4,       # update steps per same batch 3 <--> 6
                       alpha=0.35,      # termination criterion of line search, must be <= 0.35
                       beta=0.75,       # precision of line search,  imprecise 0.5 <--> 0.9 precise
                       max_step=0.02,   # similar to learning rate in SGD 0.1 <--> 0.001.
                       min_step=8e-5)

### LBFGS ($)
LBFGS_params    = dict(maxfun= 40,      # function evaluations
                       maxiter= 4,      # iterations
                       m= 10,           # maximum number of variable metric corrections
                       factr= 1e2,      # factor of machine precision as termination criterion
                       pgtol= 1e-9,     # projected gradient tolerance
                       iprint= -1)      # set to 0 for direct printing of steps
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import cPickle
import convnet
from netutils import CNNCalculator


### CNN Creation #############################################################################################
def createNet(config, input_size, n_ch, n_lab, dimension_calc):
    """
    Creates CNN according to config

    Parameters
    ----------

    n_ch: int
      Number of input channels in data
    n_lab: int
      Number of labels/classes/output_neurons
    param_file: string/path
      Optional file to initialise parameters of CNN from

    Returns
    -------

    CNN-Object
    """
    enable_dropout = False
    if len(config.dropout_rates) > 0 and (config.dropout_rates[0] is not None):
        enable_dropout = True

    recurrent = True if (config.rnn_layer_kwargs is not None) else False
    cnn = convnet.MixedConvNN(input_size,
                              input_depth=n_ch,
                              batch_size=config.batch_size,
                              enable_dropout=enable_dropout,
                              recurrent=recurrent,
                              dimension_calc=dimension_calc)

    # Conv layers
    conv = zip(config.nof_filters, config.filters, config.pool, config.activation_func, config.MFP, config.pooling_mode)
    for i, (n, f, p, act, mfp, p_m) in enumerate(conv):
        cnn.addConvLayer(n, f, p, act, use_fragment_pooling=mfp, pooling_mode=p_m)

    # RNN Layers
    if config.rnn_layer_kwargs is not None:
        cnn.addRecurrentLayer(**config.rnn_layer_kwargs)

    # MLP layers
    c = len(config.filters)
    for i, (n, act) in enumerate(zip(config.MLP_layers, config.activation_func[c:])):
        cnn.addPerceptronLayer(n, act)

        # Auto adding of last layer
    if len(config.MLP_layers) or (config.rnn_layer_kwargs is not None) > 0:
        cnn.addPerceptronLayer(n_lab, 'linear', force_no_dropout=True)
    else:
        if config.target in ['affinity', 'malis']:
            if config.target == 'malis':
                cnn.addConvLayer(n_lab, 1, 1, 'linear', is_last_layer=True, affinity='malis')  # reshape=True
            else:
                cnn.addConvLayer(n_lab, 1, 1, 'linear', is_last_layer=True, affinity=True)  # reshape=True
        else:
            cnn.addConvLayer(n_lab, 1, 1, 'linear', is_last_layer=True)  # reshape=True

    use_class_weights = True if config.class_weights is not None else False
    use_label_prop = True if config.label_prop_thresh is not None else False
    cnn.compileOutputFunctions(config.target, use_class_weights,
                               config.use_example_weights, config.lazy_labels, use_label_prop)

    cnn.setOptimizerParams(config.SGD_params, config.CG_params, config.RPROP_params,
                           config.LBFGS_params, config.Adam_params, config.weight_decay)

    if enable_dropout:
        cnn.setDropoutRates(config.dropout_rates)
    if config.param_file is not None:
        cnn.loadParameters(config.param_file)
    return cnn


def createNetfromParams(param_file,
                        patch_size,
                        batch_size=1,
                        activation_func='tanh',
                        poolings=None,
                        MFP=None,
                        only_prediction=False):
    """
    Convenience function to create CNN without ``config`` directly from a saved parameter file.
    Therefore this function only allows restricted configuration and does not initialise the training optimisers.

    Parameters
    ----------

    param_file: string/path
      File to initialise parameters of CNN from. The file must contain a list of shapes of the W-parameters as
      first entry and should ideally contain a list of pooling factors as last entry, alternatively the can be
      given as optional argument
    patch_size: tuple of int
      Patch size for input data
    batch_size: int
      Number of input patches
    activation_func: string
      Activation function to use for all layers
    poolings: list of int
      Pooling factors per layer (if not included in the parameter file)
    MFP: list of bool/{0,1}
      Whether to use MFP in the respective layers
    only_prediction: Bool
      This excludes the building of the gradient (faster)

    Returns
    -------

    CNN-Object
    """
    f = open(param_file, 'r')
    shapes = cPickle.load(f)
    n_lay = len(shapes)
    print "Shapes are:", shapes
    print "#Layers =", n_lay

    for i in xrange(n_lay):
        cPickle.load(f)
        cPickle.load(f)
    try:
        poolings = cPickle.load(f)
        print "Found Pooling info:", poolings
    except:
        if poolings is None:
            raise ValueError("No Pooling info found in file or provided")
    f.close()

    if MFP is None:
        MFP = [0, ] * n_lay

    nof_filters = map(lambda x: x[0], shapes)
    if len(shapes[0]) == 4:
        n_ch = shapes[0][1]
        filters = map(lambda x: (x[2], x[3]), shapes)
    elif len(shapes[0]) == 5:
        n_ch = shapes[0][2]
        filters = map(lambda x: (x[1], ) + x[3:], shapes)
    else:
        raise NotImplementedError()
    #filters     = map(lambda x: (x[1],)+x[3:], shapes)

    dims = CNNCalculator(filters,
                         poolings,
                         desired_input=patch_size,
                         MFP=MFP,
                         force_center=False,
                         n_dim=len(shapes[0]) - 2, )
    patch_size = dims.input

    cnn = convnet.MixedConvNN(patch_size, input_depth=n_ch, batch_size=batch_size)

    for i, (n, f, p, mfp) in enumerate(zip(nof_filters, filters, poolings, MFP)):
        if i < n_lay - 1:
            cnn.addConvLayer(n, f, p, activation_func, use_fragment_pooling=mfp)
        else:
            cnn.addConvLayer(n, 1, p, activation_func, is_last_layer=True)

    cnn.compileOutputFunctions(target='nll', only_forward=only_prediction)
    if param_file is not None:
        cnn.loadParameters(param_file)
    return cnn


if __name__ == "__main__":
    cnn = createNetfromParams('Barrier_flat_2d_end.param',
                              patch_size=(500, 500),
                              batch_size=1,
                              activation_func='tanh',
                              MFP=[1, 1, 1, 1, 0, 0, 0],
                              only_prediction=False)



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import sys
import time
from multiprocessing import Process
import numpy as np
from matplotlib import pyplot as plt

from elektronn.utils import pprinttime
from elektronn.net import introspection as intro
from elektronn.net.netcreation import createNet
import CNNData
import traindata
import trainutils
from parallelisation import BackgroundProc


class Trainer(object):
    """
    Object that manages Training of a CNN.

    Parameters
    ----------

    config: trainutils.ConfigObj
    Container for all configurations

    Examples
    --------

    All necessary configuration information is contained in cofig:

    >>> T = Trainer(config)
    >>> T.loadData()
    >>> T.createNet()
    >>> T.run() # The Training loop

    If the config options ``print_status`` and ``plot_on`` are set the CNN progress can be supervised.

    Control during iteration can be exercised by ctrl+c which evokes a commandline.
    There are various shortcuts displayed but in principle all attributes of the CNN can be accessed:

    >>> CNN MENU
    >> Debug_Run <<
    Shortcuts:
    'q' (leave interface),	    'abort' (saving params),
    'kill'(no saving),	    'save'/'load' (opt:filename),
    'sf'/' (show filters)',	    'smooth' (smooth filters),
    'sethist <int>',	    'setlr <float>',
    'setmom <float>' ,	    'params' print info,
    Change Training mode :('SGD','CG', 'RPROP', 'LBFGS')
    For EVERYTHING else enter your command in the command line
    >>> user@cnn: setlr 0.01 # Change learning rate of SGD
    >>> user@cnn: CG # Change Training mode CG (Optimizer will be compiled on demand)
    Changing Training mode...
    >>> user@cnn: self.config.savename # Access an attribute of ``trainerInstance``.
    # Inputs containing '(' or '=' will result in a print of the value
    'Debug_Run'
    >>> user@cnn: print cnn.getDropoutRates() # To see the return of function add 'print'
    [0.5, 0.5]
    >>> user@cnn: cnn.setOptimizerParams(CG={'max_step': 0.1}) # change CG-'max_step'
    >>> user@cnn: q # leave interface
    Continuing Training
    Compiling CG
      Compiling done - in 7.206 s!

    """

    def __init__(self, config=None):
        self.config = config
        self.data = None
        self.cnn = None
        self.CG_timeline = []
        self.history = []
        self.timeline = []
        self.errors = []
        self.saved_raw_preview = False

    def reset(self):
        """
        Resets all history of NLLs etc and randomizes CNN weights, optimiser hyper-parameters are set to initial
        values from config
        """
        self.cnn.randomizeWeights()
        self.cnn.setOptimizerParams(
            self.config.SGD_params, self.config.CG_params,
            self.config.RPROP_params, self.config.LBFGS_params,
            self.config.weight_decay)
        self.cnn.CG_timeline = []
        self.history = []
        self.timeline = []
        self.errors = []
        self.param_vars = []

    def run(self):
        """
        Runs the Training loop until termination. Control during iteration can be exercised by ctrl+c which
        evokes a commandline. There are various shortcuts displayed but in principle all attributes of the
        CNN can be accessed:

        Examples
        --------

        Using the command line

          >>> CNN MENU
          >> Debug_Run <<
          Shortcuts:
          'q' (leave interface),	    'abort' (saving params),
          'kill'(no saving),	    'save'/'load' (opt:filename),
          'sf'/' (show filters)',	    'smooth' (smooth filters),
          'sethist <int>',	    'setlr <float>',
          'setmom <float>' ,	    'params' print info,
          Change Training mode :('SGD','CG', 'RPROP', 'LBFGS')
          For EVERYTHING else enter your command in the command line
          >>> user@cnn: setlr 0.01 # Change learning rate of SGD
          >>> user@cnn: CG # Change Training mode CG (Optimizer will be compiled on demand)
          Changing Training mode...
          >>> user@cnn: self.config.savename # Access an attribute of ``trainerInstance``.
          # Inputs containing '(' or '=' will result in a print of the value
          'Debug_Run'
          >>> user@cnn: print cnn.getDropoutRates() # To see the return of function add 'print'
          [0.5, 0.5]
          >>> user@cnn: cnn.setOptimizerParams(CG={'max_step': 0.1}) # change CG-'max_step'
          >>> user@cnn: q # leave interface
          Continuing Training
          Compiling CG
            Compiling done - in 7.206 s!
        """
        save_name = self.config.save_name
        cnn = self.cnn
        data = self.data
        config = self.config
        schedule = self.config.LR_schedule
        t_passed = 0
        t_per_train = 1
        t_pt = 2
        t_pi = 2
        last_save_t = 0
        save_time = config.param_save_h
        last_save_t2 = 0
        save_time2 = config.initial_prev_h
        nll_ema = 0.65
        nll, train_nll, valid_nll, train_error, valid_error = 0, 0, 0, 0, 0
        user_termination = False
        plotting_proc = []
        if (schedule is not None) and (schedule != []):
            next_LR_adjust = schedule.pop(0)
        else:
            next_LR_adjust = (None, None)

        pp_loss = 'MSE' if config.target == 'regression' else 'NLL'
        pp_err = 'std' if config.target == 'regression' else '%'

        # --------------------------------------------------------------------------------------------------------
        if config.background_processes:
            n_proc = max(2, int(config.background_processes))
            bg_worker = BackgroundProc(data.getbatch, n_proc=n_proc, target_kwargs=self.get_batch_kwargs)
        # --------------------------------------------------------------------------------------------------------
        try:
            t0 = time.time()
            for i in xrange(config.n_steps):
                try:
                    if config.background_processes:
                        batch = bg_worker.get()
                    else:
                        batch = data.getbatch(**self.get_batch_kwargs)

                    if config.class_weights is not None:
                        batch = batch + (config.class_weights, )
                    if config.label_prop_thresh is not None:
                        batch = batch + (config.label_prop_thresh, )

    #-----------------------------------------------------------------------------------------------------
                    nll, nll_instance, t_per_train = cnn.trainingStep(*batch, mode=config.optimizer)  # Update step
                    #-----------------------------------------------------------------------------------------------------
                    t_per_it = time.time() - t0
                    t0 = time.time()

                    if np.any(np.isnan(nll)) or np.any(np.isinf(nll)):
                        print "The NN diverged to `nan` Loss!!!\n\
            You have the chane to inspect the last used examples and the internal state of pipeline in the\
            command line. The last presented training input data is `batch[0]` and the corresponding target `batch[1]`"

                        raise KeyboardInterrupt

                    nll_ema = 0.995 * nll_ema + 0.005 * nll  # EMA
                    t_pt = 0.8 * t_pt + 0.2 * t_per_train  # EMA
                    t_pi = 0.8 * t_pi + 0.2 * t_per_it  # EMA
                    t_passed += t_per_it
                    batch_char = batch[1].mean()
                    self.timeline.append([i, t_passed, nll_ema, nll, batch_char])
                    if (t_passed - last_save_t) / 3600 > config.param_save_h:  # every hour
                        last_save_t = t_passed
                        time_string = '-' + str(save_time) + 'h'
                        cnn.saveParameters(save_name + time_string + '.param', show=False)
                        save_time += config.param_save_h

                    if self.preview_data is not None:
                        if (t_passed-last_save_t2)/3600 > config.prev_save_h or (t_passed/3600 > config.initial_prev_h and last_save_t2==0): # first time
                            last_save_t2 = t_passed
                            config.preview_kwargs['number'] = save_time2
                            save_time2 += config.prev_save_h
                            try:
                                self.previewSlice(**config.preview_kwargs)
                            except:
                                print "Preview Predictions failed. Are the preview raw data in the correct format?"

                    if i == next_LR_adjust[0]:
                        cnn.setSGDLR(np.float32(next_LR_adjust[1]))
                        try:
                            next_LR_adjust = schedule.pop(0)
                        except IndexError:  # list is empty
                            next_LR_adjust = (None, None)

                    if i % 1000 == 0:  # update learning rate (exp. decay)
                        cnn.setSGDLR(np.float32(cnn.SGD_LR.get_value() * config.LR_decay))

                    if (i % config.history_freq[0] == 0) and config.history_freq[0] != 0:
                        lr = cnn.SGD_LR.get_value()
                        self.CG_timeline = cnn.CG_timeline
                        ### Training  & Valid Errors ###
                        nll_after = cnn.get_loss(*batch)[0]
                        nll_gain = nll_after - nll
                        train_nll, train_error = self.testModel('train')
                        valid_nll, valid_error = self.testModel('valid')
                        if config.target != 'regression':
                            train_error *= 100
                            valid_error *= 100

                        self.errors.append([i, t_passed, train_error, valid_error])
                        self.history.append([i, t_passed, nll_ema, nll, train_nll, valid_nll, nll_gain, lr])
                        if config.target == 'malis':
                            self.malisPreviewSlice(batch, name=i)

                            ### Monitoring / Output ###
                            #np.save('Backup/'+save_name+".DataHist",  np.array(self.data.HIST)) ### DEBUG
                        cnn.saveParameters(save_name + '-LAST.param', show=False)

                        if config.plot_on and i > 30:
                            [p.join() for p in plotting_proc]  # join before new plottings are started
                            plotting_proc = []
                            p0 = Process(target=trainutils.plotInfo,
                                         args=(self.timeline, self.history, self.CG_timeline, self.errors, save_name))
                            plotting_proc.extend([p0])
                            [p.start() for p in plotting_proc]
                        else:
                            trainutils.saveHist(self.timeline, self.history, self.CG_timeline, self.errors, save_name)

                        if config.print_status:
                            out = '%05i %sm=%.3f, train=%05.2f%s, valid=%05.3f%s, prev=%04.1f, NLLdiff=%+.1e, LR=%.5f, %.1f it/s, ' % (i, pp_loss, nll_ema, train_error, pp_err, valid_error, pp_err, batch_char*100, nll_gain, cnn.SGD_LR.get_value(), 1.0 / t_pi)
                            t = pprinttime(t_passed)
                            print out + t

    # User Interface #####################################################################################
                except KeyboardInterrupt:
                    out = '%05i %s=%.5f, NLL=%.4f, train=%.5f, valid=%.5f, train=%.3f%s, valid=%.3f%s,\n\
      LR=%.6f, MOM=%.6f, %.1f GPU-it/s, %.1f CPU-it/s, '\
                    % (i, pp_loss, nll_ema, nll, train_nll, valid_nll, train_error, pp_err, valid_error, pp_err,
                      cnn.SGD_LR.get_value(), cnn.SGD_momentum.get_value(),1.0/t_pt, 1.0/t_pi)

                    t = pprinttime(t_passed)
                    print out + t

                    # Like a command line, it must be here to access workspace variables
                    trainutils.pprintmenu(save_name)
                    while True:
                        try:
                            ret = trainutils.userInput(cnn, config.history_freq)
                            plt.pause(0.001)
                            if ret is None or ret == "":
                                continue
                            if ret == "abort":
                                user_termination = True
                                break
                            elif ret == 'kill':
                                return
                            elif ret in ['SGD', 'RPROP', 'CG', 'LBFGS', 'Adam']:
                                config.optimizer = ret
                            elif ret == 'q':
                                print "Continuing Training"
                                break
                            elif ret == 'sf':
                                intro.plotFilters(cnn)
                            else:
                                if '(' in ret or '=' in ret:  # execute statements and assignments
                                    exec(ret)
                                else:  # print value of identifiers
                                    exec('print ' + ret)

                        except:
                            sys.excepthook(*sys.exc_info())  # show info on error

                    if self.config.background_processes:
                        bg_worker.reset()

                    t0 = time.time()  # reset time after user interaction, otherwise time will appear as pause in plot

    # End UI ###############################################################################################
                if (t_passed > config.max_runtime) or user_termination:  # This is in the epoch/UI loop
                    print 'Timeout or manual Termination'
                    break
            # This is OUTSIDE the training loop i.e. the last block of the function ``run``
            self.cnn.saveParameters(save_name + "_end.param")
            trainutils.plotInfo(self.timeline, self.history, self.CG_timeline, self.errors, save_name)
            print 'End of Training'
            print '#' * 60 + '\n' + '#' * 60 + '\n'
            # -------------------end of run()---------------------------------------------------------------------------

        except:
            sys.excepthook(*sys.exc_info())  # show info on error
        finally:
            if config.background_processes:
                bg_worker.shutdown()

    def loadData(self):
        config = self.config
        if self.config.mode != 'vect-scalar' and self.config.data_class_name is None:  # image training
            strided = ~np.any(config.MFP) and config.mode == 'img-img'

            self.get_batch_kwargs = dict(
                batch_size=config.batch_size,
                strided=strided,
                flip=config.flip_data,
                grey_augment_channels=config.grey_augment_channels,
                ret_info=config.lazy_labels,
                ret_example_weights=config.use_example_weights,
                warp_on=config.warp_on,
                ignore_thresh=config.example_ignore_threshold)

            # the source is replaced in self.testModel to be valid
            self.get_batch_kwargs_test = dict(
                batch_size=config.monitor_batch_size,
                strided=strided,
                flip=config.flip_data,
                grey_augment_channels=config.grey_augment_channels,
                ret_info=config.lazy_labels,
                ret_example_weights=config.use_example_weights,
                warp_on=False,
                ignore_thresh=config.example_ignore_threshold)  # no warp

            self.data = CNNData.CNNData(
                config.patch_size, config.dimensions.pred_stride,
                config.dimensions.offset, config.n_dim, config.n_lab,
                config.anisotropic_data, config.mode, config.zchxy_order,
                config.border_mode, config.pre_process, config.upright_x, True
                if config.target == 'regression' else False, config.target
                if config.target in ['malis', 'affinity'] else False)  # return affinity graph instead of boundaries

            self.data.addDataFromFile(config.data_path, config.label_path,
                                      config.d_files, config.l_files,
                                      config.cube_prios, config.valid_cubes,
                                      config.downsample_xy)

            if self.config.preview_data_path is not None:
                data = trainutils.h5Load(self.config.preview_data_path)
                if not (isinstance(data, list) or isinstance(data, (tuple, list))):
                    #data = np.transpose(data, (1,2,0)) # this was only a hack for I
                    data = [data, ]

                data = [d.astype(np.float32) / 255 for d in data]
                self.preview_data = data
            else:
                self.preview_data = None

        else:  # non-image training
            self.get_batch_kwargs = dict(batch_size=config.batch_size)
            self.get_batch_kwargs.update(self.config.data_batch_kwargs)
            # the source is replaced in self.testModel to be valid
            self.get_batch_kwargs_test = dict(batch_size=config.monitor_batch_size)
            if isinstance(self.config.data_class_name, tuple):
                Data = trainutils.import_variable_from_file(*self.config.data_class_name)
            else:
                Data = getattr(traindata, self.config.data_class_name)

            self.data = Data(**self.config.data_load_kwargs)
            self.preview_data = None

    def createNet(self):
        """
        Creates CNN according to config
        """
        n_lab = self.data.n_lab
        if self.config.class_weights is not None:
            if self.config.target == 'nll':
                assert len(self.config.class_weights) == n_lab,\
                       "The number of class weights must equal the number of classes"

        if self.config.mode != 'vect-scalar':  # image training
            n_ch = self.data.n_ch
            self.cnn = createNet(self.config, self.config.patch_size, n_ch, n_lab, self.config.dimensions)
        else:  # non-image training
            n_ch = None
            if self.config.rnn_layer_kwargs is not None:
                n_ch = self.data.n_taps  # must be None if the data should be repeated

            self.cnn = createNet(self.config, self.data.example_shape, n_ch, n_lab, None)

    def debugGetCNNBatch(self):
        """
        Executes ``getbatch`` but with un-strided labels and always returning info. The first batch example
        is plotted and the whole batch is returned for inspection.
        """
        if self.config.mode == 'img-img':
            batch = self.data.getbatch(
                self.config.monitor_batch_size,
                source='train',
                strided=False,
                flip=self.config.flip_data,
                grey_augment_channels=self.config.grey_augment_channels,
                ret_info=True)

            try:
                data, label, info1, info2 = batch
            except:
                data, label, seg, info1, info2 = batch

            if len(label.shape) == 5:  # affinities (bs, 3, z, x, y)
                print "Plot Batch: Showing min affinity only."
                label = np.min(label, axis=1)

            if self.config.n_dim == 2:
                CNNData.plotTrainingTarget(data[0, 0], label[0], 1)
            else:
                i = int(self.config.dimensions.offset[0])
                CNNData.plotTrainingTarget(data[0, i, 0], label[0, 0], 1)
#      print "Info1=",info1
#      print "Info2=",info2
            plt.show()
            plt.savefig('debugGetCNNBatch.png', bbox_inches='tight')
            plt.pause(0.01)
            plt.pause(2.0)
            plt.close('all')
            plt.pause(0.01)

            return data, label, info1, info2
        else:
            print "This funciton is only available for 'img-img' training mode"

    def testModel(self, data_source):
        """
        Computes NLL and error/accuracy on batch with ``monitor_batch_size``


        Parameters
        ----------

        data_source: string
            'train' or 'valid'

        Returns
        -------
        NLL, error:

        """
        if data_source == 'valid':
            if not hasattr(self.data, 'valid_d') or not hasattr(self.data.valid_d, '__len__') or len(self.data.valid_d) == 0:
                return np.nan, np.nan  # 0, 0

        kwargs = dict(self.get_batch_kwargs_test)  # copy because it is modified in next line!
        kwargs['source'] = data_source
        batch = self.data.getbatch(**kwargs)

        y_aux = []
        if self.config.class_weights is not None:
            y_aux.append(self.config.class_weights)

        if self.config.label_prop_thresh is not None:
            y_aux.append(self.config.label_prop_thresh)

        rates = self.cnn.getDropoutRates()
        self.cnn.setDropoutRates([0.0, ] * len(rates))
        n = len(batch[0])
        nll = 0
        error = 0
        for j in xrange(int(np.ceil(np.float(n) / self.config.batch_size))):
            d = batch[0][j * self.config.batch_size:(j + 1) * self.config.batch_size]  # data
            l = batch[1][j * self.config.batch_size:(j + 1) * self.config.batch_size]  # label
            if len(batch) > 2:
                aux = []
                for b in batch[2:]:
                    aux.append(b[j * self.config.batch_size:(j + 1) * self.config.batch_size])

                nl, er, pred = self.cnn.get_error(d, l, *(aux + y_aux))
            else:
                nl, er, pred = self.cnn.get_error(d, l, *y_aux)
            nll += nl * len(d)
            error += er * len(d)
        nll /= n
        error /= n
        self.cnn.setDropoutRates(rates)  # restore old rates
        return nll, error

    def predictAndWrite(self, raw_img, number=0, export_class='all', block_name='', z_thick=5):
        """
        Predict and and save a slice as preview image

        Parameters
        ----------

        raw_img : np.ndarray
          raw data in the format (ch, x, y, z)
        number: int/float
          consecutive number for the save name (i.e. hours, iterations etc.)
        export_class: str or int
          'all' writes images of all classes, otherwise only the class with index ``export_class`` (int) is saved.
        block_name: str
          Name/number to distinguish different raw_imges
        """

        block_name = str(block_name)
        pred = self.cnn.predictDense(raw_img)  # returns (k, x, y(, z))

        z_sh = pred.shape[-1]

        if pred.shape[0] == 3:
            print "WARNING: hack active for affinity previews"
            pred[0] = pred.min(axis=0)

        pred = pred[:, :, :, (z_sh - z_thick) // 2:(z_sh - z_thick) // 2 + z_thick]

        save_name = self.config.save_name
        for z in xrange(pred.shape[3]):
            if export_class == 'all':
                for c in xrange(pred.shape[0]):
                    plt.imsave('%s-pred-%s-c%i-z%i-%shrs.png' % (save_name, block_name, c, z, number), pred[c,:,:,z], cmap='gray')
            elif export_class in ['malis', 'affinity']:
                plt.imsave('%s-pred-%s-aff-z%i-%shrs.png' % (save_name, block_name, z, number),
                np.transpose(pred[0:6:2,:,:,z],(1,2,0)), cmap='gray')
            else:
                if isinstance(export_class, (list, tuple)):
                    for c in export_class:
                        plt.imsave('%s-pred-%s-c%i-z%i-%shrs.png' % (save_name, block_name, c, z, number), pred[c,:,:,z], cmap='gray')

                else:
                    c = int(export_class)
                    plt.imsave('%s-pred-%s-c%i-z%i-%shrs.png' % (save_name, block_name, c, z, number), pred[c,:,:,z], cmap='gray')

        if not self.saved_raw_preview:  # only do once
            z_off = 0 if len(self.config.dimensions.offset) == 2 else int(self.config.dimensions.offset[0])
            for z in xrange(pred.shape[3]):
                plt.imsave('%s-raw-%s-z%i.png' % (save_name, block_name, z), raw_img[0, :, :, z + z_off], cmap='gray')

    def previewSliceFromTrainData(self, cube_i=0, off=(0, 0, 0), sh=(10, 400, 400), number=0, export_class='all'):
        """
        Predict and and save a selected slice from the training data as preview

        Parameters
        ----------

        cube_i: int
          index of source cube in CNNData
        off: 3-tuple of int
          start index of slice to cut from cube (z,x,y)
        sh: 3-tuple of int
          shape of cube to cut (z,x,y)
        number: int
          consecutive number for the save name (i.e. hours, iterations etc.)
        export_class: str or int
          'all' writes images of all classes, otherwise only the class with index ``export_class`` (int) is saved.
        """
        if not self.config.mode == 'img-img':
            print "This funciton is only available for 'img-img' training mode"
            return

        if self.cnn.n_dim == 3:
            min_z = self.cnn.input_shape[1]
            if min_z > sh[0]:
                sh = list(sh)
                sh[0] = min_z

        raw_img = self.data.train_d[cube_i]
        raw_img = raw_img[off[0]:off[0] + sh[0], :, off[1]:off[1] + sh[1], off[1]:off[1] + sh[1]]

        raw_img = np.transpose(raw_img, (1, 2, 3, 0))  # (z,ch,x,y) --> (ch,x,y,z)
        self.predictAndWrite(raw_img, number, export_class)
        self.saved_raw_preview = True

    def previewSlice(self, number=0, export_class='all', max_z_pred=5):
        """
        Predict and and save a data from a separately loaded file as preview

        Parameters
        ----------

        number: int/float
          consecutive number for the save name (i.e. hours, iterations etc.)
        export_class: str or int
          'all' writes images of all classes, otherwise only the class with index ``export_class`` (int) is saved.
        max_z_pred: int
          approximate maximal number of z-slices to produce (depends on CNN architecture)
        """
        if not self.config.mode == 'img-img':
            print "This funciton is only available for 'img-img' training mode"
            return

        assert self.preview_data is not None, "You must provide preview data in order to call this function"
        for example_no, raw_img in enumerate(self.preview_data):
            z_sh = raw_img.shape[-1]
            if self.cnn.n_dim == 3:
                strd_z = self.cnn.output_strides[0]
                out_z = self.cnn.output_shape[2] * strd_z
                min_z = self.cnn.input_shape[1] + strd_z - 1
                z_thick = min_z if out_z > max_z_pred else min_z + strd_z * int(np.ceil(float(max_z_pred - out_z) / strd_z))
            else:
                z_thick = max_z_pred

            assert z_thick <= z_sh, "The preview slices are too small in z-direction for this CNN"

            if raw_img.ndim == 3:
                raw_img = raw_img[None, :, :, (z_sh - z_thick) // 2:(z_sh - z_thick) // 2 + z_thick]
            elif raw_img.ndim == 4:
                raw_img = raw_img[:, :, :, (z_sh - z_thick) // 2:(z_sh - z_thick) // 2 + z_thick]

            self.predictAndWrite(raw_img, number, export_class, example_no, max_z_pred)

        self.saved_raw_preview = True

    def malisPreviewSlice(self, batch, name='A'):
        pred = self.cnn.class_probabilities(batch[0])[0]  # (6, z, x,y)
        malis = self.cnn.malis_stats(*batch[:3])  # nll, n_pos, n_neg, n_tot, false_splits, false_merges, rand_index, pos_count, neg_count, labels
        nll, n_pos, n_neg, n_tot, false_splits, false_merges, rand_index, pos_count, neg_count = malis
        data, aff_gt, seg_gt = batch[:3]

        print "NLL    : ", nll
        print "N total: ", n_tot
        print "N pos  : ", n_pos
        print "N neg  : ", n_neg
        print "Splits : ", false_splits
        print "Mergers: ", false_merges
        print "Rand-Index: ", rand_index

        pred_slices = np.transpose(pred[1::2], (1, 2, 3, 0))
        pos_slices = np.transpose(pos_count, (1, 2, 3, 0))
        neg_slices = np.transpose(neg_count, (1, 2, 3, 0))
        neg_slices = np.log(neg_slices + 1)
        data = data[0, :, 0]
        aff_gt = np.transpose(aff_gt[0], (1, 2, 3, 0))
        seg_gt = seg_gt[0]

        trainutils.pickleSave([pred_slices, aff_gt, pos_slices, neg_slices, seg_gt, data], 'MALIS-%s.pkl' % (name, ))



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import numpy as np
import time
import theano
import theano.tensor as T
from theano.tensor.shared_randomstreams import RandomStreams

import pooling
from netutils import initWeights


class PerceptronLayer(object):
    """
    Typical hidden layer of a MLP: units are fully-connected.
    Weight matrix W is of shape (n_in,n_out), the bias vector b is of shape (n_out,).

    :type input: theano.tensor.dmatrix
    :param input: a symbolic tensor of shape (n_examples, n_in)

    :type n_in: int
    :param n_in: dimensionality of input

    :type n_out: int
    :param n_out: number of hidden units

    :type batch_size: int
    :param batch_size: batch_size

    :type enable_dropout: Bool
    :param pool: whether to enable dropout in this layer. The default rate is 0.5 but it can be changed with
               self.activation_noise.set_value(set_value(np.float32(p)) or using cnn.setDropoutRates

    :type activation_func: string
    :param activation_func: {'relu','sigmoid','tanh','abs', 'maxout <i>'}

    :type input_noise: theano.shared float32
    :param input_noise: std of gaussian (centered) input noise. 0 or None --> no noise

    :type input_layer: layer object
    :param input_layer: just for keeping track of un-usual input layers

    :type W: np.ndarray or T.TensorVariable
    :param W: weight matrix. If array, the values are used to initialise a shared variable for this layer.
                           If TensorVariable, than this variable is directly used (weight sharing with the
                           layer from which this variable comes from)

    :type b: np.ndarray or T.TensorVariable
    :param b: bias vector. If array, the values are used to initialise a shared variable for this layer.
                           If TensorVariable, than this variable is directly used (weight sharing with the
                           layer from which this variable comes from)
    """

    def __init__(self,
                 input,
                 n_in,
                 n_out,
                 batch_size,
                 enable_dropout,
                 activation_func='tanh',
                 input_noise=None,
                 input_layer=None,
                 W=None,
                 b=None):
        self.input_layer = input_layer  # only for autoencoder
        self.activation_func = activation_func
        self.output_shape = (batch_size, n_out)
        self.n_in = n_in
        self.lin_output = None
        self.output = None
        self.last_grads = []  # only for autoencoder

        print "PerceptronLayer( #Inputs =", n_in, "#Outputs =", n_out, ")"
        if input_noise is not None:
            self.input_noise = theano.shared(np.float32(input_noise), name='Input Noise')
            print "Input_noise active, p=" + str(self.input_noise.get_value())
            rng = np.random.RandomState(int(time.time()))
            theano_rng = RandomStreams(rng.randint(2**30))
            # apply multiplicative noise to input
            #self.input = theano_rng.binomial(size=input.shape, n=1, p=1-self.input_noise,
            #                                                                dtype=theano.config.floatX) * input
            # apply additive noise to input
            self.input = input + theano_rng.normal(size=input.shape,
                                                   avg=0,
                                                   std=input_noise,
                                                   dtype=theano.config.floatX)
        else:  # no input noise
            self.input = input

        if W is None:
            W_values = np.asarray(initWeights((n_in, n_out), scale='glorot', mode='uni'), dtype=theano.config.floatX)
            self.W = theano.shared(value=W_values, name='W_perceptron' + str(n_in) + '.' + str(n_out), borrow=True)
        else:
            print "Directly using fixed/shared W (", W, "), no Training on it in this layer!"
            if isinstance(W, np.ndarray):
                self.W = theano.shared(value=W.astype(np.float32),
                                       name='W_perceptron' + str(n_in) + '.' + str(n_out),
                                       borrow=True)
            else:
                assert isinstance(W, T.TensorVariable), "W must be either np.ndarray or theano var"
                self.W = W

        if b is None:
            #b_values = np.asarray(np.random.uniform(-1e-8,1e-8,(n_out,)), dtype=theano.config.floatX)
            if activation_func == 'relu' or activation_func == 'ReLU':
                b_values = np.asarray(initWeights((n_out, ), scale=1.0, mode='const'), dtype=theano.config.floatX)
            elif activation_func == 'sigmoid':
                b_values = np.asarray(initWeights((n_out, ), scale=0.5, mode='const'), dtype=theano.config.floatX)
            else:  # activation_func=='tanh':
                b_values = np.asarray(initWeights((n_out, ), scale=1e-6, mode='fix-uni'), dtype=theano.config.floatX)

            self.b = theano.shared(value=b_values, name='b_perceptron' + str(n_in) + '.' + str(n_out), borrow=True)

        else:
            print "Directly using fixed given b (", b, "), no Training on it in this layer!"
            if isinstance(b, np.ndarray):
                self.b = theano.shared(value=b.astype(np.float32),
                                       name='b_perceptron' + str(n_in) + '.' + str(n_out),
                                       borrow=True)
            else:
                assert isinstance(b, T.TensorVariable), "b must be either np.ndarray or theano var"
                self.b = b

        lin_output = T.dot(self.input, self.W)

        if enable_dropout:
            print "Dropout ON"
            self.activation_noise = theano.shared(np.float32(0.5), name='Dropout Rate')
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p = 1 - self.activation_noise
            self.dropout_gate = 1.0 / p * rng.binomial((n_out, ), 1, p, dtype=theano.config.floatX)
            lin_output = lin_output * self.dropout_gate.dimshuffle(('x', 0))

        lin_output = lin_output + self.b

        # Apply non-linearities and ggf. change bias-initialisations
        if activation_func == 'tanh':  # range = [-1,1]
            self.output = T.tanh(lin_output)  # shape: (batch_size, num_outputs)
        elif activation_func == 'relu' or activation_func == 'ReLU':  # rectified linear unit ,range = [0,inf]
            self.activation_func = 'relu'
            self.output = lin_output * (lin_output > 0)  #T.maximum(lin_output,T.zeros_like(lin_output))
        elif activation_func == 'abs':  # abs unit ,range = [0,inf]
            self.output = T.abs_(lin_output)
        elif activation_func == 'sigmoid':  # range = [0,1]
            #print "WARNING: consider using tanh(.) or relu(.) instead! Sigmoid is BAD! (relu > tanh >> sigmoid)"
            lin_output = T.dot(self.input, self.W) + self.b
            self.output = T.nnet.sigmoid(lin_output)  #1/(1 + T.exp(-lin_output))
        elif activation_func == 'linear':
            self.output = (lin_output)
        elif activation_func.startswith("maxout"):
            r = int(activation_func.split(" ")[1])
            assert r >= 2
            n_out = n_out / r
            self.output = pooling.maxout(lin_output, factor=r)
        else:
            raise NotImplementedError("Options are: activation_func=('relu'|'sigmoid'|'tanh'|'abs')")

        self.lin_output = lin_output
        self.params = [self.b if b is None else []] + ([self.W]
                                                       if W is None else [])
        self.class_probabilities = T.nnet.softmax(lin_output)  # shape: (batch_size, num_outputs)
        #self.class_probabilities = T.exp(lin_output) / T.sum(T.exp(lin_output), axis=1, keepdims=True) # For Hessian
        self.class_prediction = T.argmax(self.class_probabilities, axis=1)  # shape: (batch_size,)

    #############################################################################################

    def randomizeWeights(self, scale='glorot', mode='uni'):
        n_in = self.n_in
        n_out = self.output_shape[1]
        if self.activation_func == 'relu':
            b_values = np.asarray(initWeights((n_out, ), scale=1.0, mode='const'), dtype=theano.config.floatX)
        elif self.activation_func == 'sigmoid':
            b_values = np.asarray(initWeights((n_out, ), scale=0.5, mode='const'), dtype=theano.config.floatX)
        else:  #self.activation_func=='tanh':
            b_values = np.asarray(initWeights((n_out, ), scale=1e-6, mode='fix-uni'), dtype=theano.config.floatX)

        W_values = np.asarray(initWeights((n_in, n_out), scale, mode), dtype=theano.config.floatX)

        self.W.set_value(W_values)
        self.b.set_value(b_values)

    def NLL(self, y, class_weights=None, example_weights=None, label_prop_thresh=None):
        """
        Returns the symbolic mean and instance-wise negative log-likelihood of the prediction
        of this model under a given target distribution.

        y: theano.tensor.TensorType
          corresponds to a vector that gives for each example the correct label. Labels < 0 are ignored (e.g. can
          be used for label propagation)

        class_weights: theano.tensor.TensorType
          weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
          ``w > 1.0`` (more important class)

        label_prop_thresh: float (0.5,1)
          This threshold allows unsupervised label propagation (only for examples with negative/ignore labels).
          If the predictive probability of the most likely class exceeds the threshold, this class is assumed to
          be the correct label and the training is pushed in this direction.
          Should only be used with pre-trained networks, and values <= 0.5 are disabled.
        """

        # NOTE: This whole function has a ugly problem with NaN. They arise for pred values close to 0 or 1
        # (i.e. for NNs that make very confident and usually also correct predictions) because initially the log of
        # all the whole pred tensor is taken. Later we want to use only some indices of the tensor (mask) but
        # that is not so easy done. There are two ways:
        # 1. advanced indexing: T.log(pred)[mask.nonzero()] --> fails if mask is all zero, cannot be fixed
        # 2. multiplying with 0-1-mask: T.log(pred) * mask.nonzero --> but NaN * 0 = NaN, but we require 0!
        # For the second option, in principle, the NaNs could be replaced by 0 using T.switch, but then the gradient
        # fails because the replaced value is disconnected from the parameters and gives NaN (mathematically
        # the gradient should correctly be 0 then; there is a Theano ticket open to request a fix).
        # So finally the best practice is to add a stabilisation to the log: T.log(pred) --> T.log(pred+eps)
        # This looks ugly, but does the task and the introduced error is completely negligible
        eps = 1e-6
        pred = self.class_probabilities  # predictive (bs, cl)
        y = y.dimshuffle(0, 'x')  # the labels (bs, 1)
        cls = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0)  # available classes
        label_selection = T.eq(cls, y)  # selects correct labels

        if class_weights is None:
            class_weights = T.ones_like(pred)
        else:
            class_weights = class_weights.dimshuffle('x', 0)

            # Up vote block
        nll_inst_up = -T.log(pred + eps) * label_selection * class_weights
        N_up = T.sum(label_selection)  # number of labelled examples

        if label_prop_thresh is not None:  # Label propagation block
            above_thresh = pred > label_prop_thresh  # this is one for the class with highes prob
            prop_mask = above_thresh * (1 - label_selection.sum(axis=1))  # don't do where training labels are available
            nll_inst_up_prop = -T.log(pred + pred) * prop_mask * class_weights
            N_up_prop = prop_mask.sum()

            nll_inst_up += nll_inst_up_prop
            N_up += N_up_prop

        nll_inst = nll_inst_up
        N_up = T.switch(T.eq(N_up, 0), 1, N_up)  # patch N to be not 0, when this is the case the sum is 0 anyway!
        nll = nll_inst.sum() / N_up

        return nll, nll_inst

    def NLL_weak(self,
                 y,
                 class_weights=None,
                 example_weights=None,
                 label_prop_thresh=None):
        """
        Returns the symbolic mean and instance-wise negative log-likelihood of the prediction
        of this model under a given target distribution.

        y: theano.tensor.TensorType
          corresponds to a vector that gives for each example the correct label. Labels < 0 are ignored (e.g. can
          be used for label propagation)

        class_weights: theano.tensor.TensorType
          weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
          ``w > 1.0`` (more important class)

        label_prop_thresh: float (0.5,1)
          This threshold allows unsupervised label propagation (only for examples with negative/ignore labels).
          If the predictive probability of the most likely class exceeds the threshold, this class is assumed to
          be the correct label and the training is pushed in this direction.
          Should only be used with pre-trained networks, and values <= 0.5 are disabled.
        """

        # NOTE: This whole function has a ugly problem with NaN. They arise for pred values close to 0 or 1
        # (i.e. for NNs that make very confident and usually also correct predictions) because initially the log of
        # all the whole pred tensor is taken. Later we want to use only some indices of the tensor (mask) but
        # that is not so easy done. There are two ways:
        # 1. advanced indexing: T.log(pred)[mask.nonzero()] --> fails if mask is all zero, cannot be fixed
        # 2. mutiplying with 0-1-mask: T.log(pred) * mask.nonzero --> but NaN * 0 = NaN, but we require 0!
        # For the second option, in principle, the NaNs could be replaced by 0 using T.switch, but then the gradient
        # fails because the replaced value is disconnected from the parameters and gives NaN (mathematically
        # the gradient should correctly be 0 then; there is a Theano ticket open to request a fix).
        # So finally the best practice is to add a stabilisation to the log: T.log(pred) --> T.log(pred+eps)
        # This looks ugly, but does the task and the introduced error is completely negligible
        eps = 1e-6
        pred = self.class_probabilities  # predictive (bs, cl)
        y = y.dimshuffle(0, 'x')  # the labels (bs, 1)
        cls = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0)  # available classes
        hard_labels = T.eq(cls, y)  # selects correct labels

        if class_weights is None:
            class_weights = T.ones_like(pred)
        else:
            class_weights = class_weights.dimshuffle('x', 0)

        soft_labels = 0.5 * hard_labels + 0.5 * pred

        # Up vote block
        nll_inst_up = -(soft_labels * T.log(pred + eps)) * class_weights
        N_up = T.sum(hard_labels)  # number of labelled examples

        nll_inst = nll_inst_up
        N_up = T.switch(T.eq(N_up, 0), 1, N_up)  # patch N to be not 0, when this is the case the sum is 0 anyway!
        nll = nll_inst.sum() / N_up

        return nll, nll_inst

    def nll_mutiple_binary(self, y, class_weights=None):
        """
        Returns the mean and instance-wise negative log-likelihood of the prediction
        of this model under a given target distribution.

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label

        Note: we use the mean instead of the sum so that
              the learning rate is less dependent on the batch size
        """
        # y (bs, n_lab)
        eps = 1e-6
        act = self.lin_output  # (bs, n_lab)
        prob_0 = T.exp(act) / (T.exp(act) + 1)
        prob_1 = 1.0 - prob_0

        self.class_probabilities = T.stack(prob_0, prob_1).dimshuffle(1, 2, 0)  # (bs, n_lab, 2)
        self.class_prediction = T.argmax(self.class_probabilities, axis=2)
        if class_weights is None:
            class_weights = T.ones(2)
        else:
            class_weights = class_weights

        nll_inst = (-T.log(prob_0 + eps) * (1 - y) * class_weights[0] - T.log(prob_1 + eps) * y * class_weights[1])
        nll = T.mean(nll_inst)
        return nll, nll_inst

    def squared_distance(self, Target, Mask=None, return_instancewise=True):
        """
        Target is the TARGET image (vectorized), -> shape(x) = (batchsize, n_target)
        output: scalar float32
        mask: vectorized, 1==hole, 0==no_hole (== DOES NOT TRAIN ON NON-HOLES)
        """
        if Mask is None:
            batch = T.mean((self.output - Target)**2)
            inst = (self.output - Target)**2
        else:
            print "squared_distance::Masked"
            #batch = T.mean(((self.output - Target)*T.concatenate( (Mask,Mask,Mask),axis=1 )  )**2 ) #assuming RBG input
            batch = T.mean(((self.output - Target) * Mask)**2)
            inst = ((self.output - Target) * Mask)**2
        if return_instancewise:
            return batch, inst
        else:
            return batch

    def errors(self, y):
        """
        Returns classification accuracy

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label
        """
        # check if y has same dimension of class_prediction
        if y.ndim != self.class_prediction.ndim:
            raise TypeError('y should have the same shape as self.class_prediction',
                            ('y', y.type, 'class_prediction', self.class_prediction.type))
        # check if y is of the correct datatype
        if y.dtype.startswith('int'):
            # the T.neq operator returns a vector of 0s and 1s, where 1
            # represents a mistake in prediction
            return T.mean(T.neq(self.class_prediction, y))
        else:
            raise NotImplementedError()

    def errors_no_tn(self, y):
        pred = self.class_prediction
        tp = T.sum((y * pred))
        #tn = T.sum((1-y)*(1-pred))
        fp = T.sum((1 - y) * pred)
        fn = T.sum(y * (1 - pred))
        acc = tp.astype('float32') / (tp + fp + fn)
        return acc

    def _make_window(self):
        print "window is on 32x32, fixed sigma, assuming RGB."
        denom = 29.8
        x0 = 16
        sig = 19
        fun = lambda z, x, y: (32 / denom * np.exp(-(abs(x - x0))**3 / (2 * sig**3))) * (32 / denom * np.exp(-(abs(y - x0))**3 / (2 * sig**3)))  #, {x, 0, 32}, {y, 0, 32}
        return np.fromfunction(fun, (3, 32, 32))

    def cross_entropy_array(self, Target, Mask=None, GaussianWindow=False):
        """
        Target is the TARGET image (vectorized), -> shape(x) = (batchsize, imgsize**2)
        the output is of length: <batchsize>, Use cross_entropy() to get a scalar output.
        """
        if GaussianWindow:
            window = self.__make_window().reshape(1, -1)
        if Mask is None:
            #XX = window#T.TensorConstant(T.TensorType(theano.config.floatX,[True,False])(),data=window)
            return -T.mean(
                (1. if GaussianWindow == False else window) *
                (T.log(self.class_probabilities) * Target +
                 T.log(1.0 - self.class_probabilities) * (1.0 - Target)),
                axis=1)
        else:
            return -T.mean(
                (T.log(self.class_probabilities) * Target +
                 T.log(1.0 - self.class_probabilities) *
                 (1.0 - Target)) * T.concatenate(
                     (Mask, Mask, Mask),
                     axis=1),
                axis=1)  #assuming RBG input

    def cross_entropy(self, Target, Mask=None, GaussianWindow=False):
        """
        Target is the TARGET image (vectorized), -> shape(x) = (batchsize, imgsize**2) output: scalar float32
        """
        if GaussianWindow:
            window = self._make_window().reshape(1, -1)
        if Mask is None:
            #XX = window#T.TensorConstant(T.TensorType(theano.config.floatX,[True,False])(),data=window)
            return -T.mean(
                (1. if GaussianWindow == False else window) *
                (T.log(self.class_probabilities) * Target +
                 T.log(1.0 - self.class_probabilities) * (1.0 - Target))
            )  # #.reshape(new_shape)[index[0]:index[2],index[1]:index[3]]
        else:
            print "cross_entropy::Masked, no window"
            return -T.mean(
                (T.log(self.class_probabilities) * Target +
                 T.log(1.0 - self.class_probabilities) *
                 (1.0 - Target)) * T.concatenate(
                     (Mask, Mask, Mask),
                     axis=1)
            )  # #.reshape(new_shape)[index[0]:index[2],index[1]:index[3]]#assuming RBG input


class RecurrentLayer(object):
    """
    :type input: symbolic input carrying [time, batch, feat]
    :param input: theano.tensor.ftensor3

    :type n_in: int
    :param n_in: dimensionality of input

    :type n_hid: int
    :param n_hid: number of hidden units

    :type activation_func: string
    :param activation_func: {'relu','sigmoid','tanh','abs'}
    """

    def __init__(self, input, n_in, n_hid, batch_size, activation_func='tanh'):
        assert input.ndim == 3
        input = input.dimshuffle(1, 0, 2)  # exchange batch and time
        self.n_in = n_in
        self.n_hid = n_hid
        self.activation_func = activation_func
        self.output_shape = (batch_size, n_hid)
        self.hid_lin = None
        self.output = None

        print "RecurrentLayer( #Inputs =", n_in, "#Hidden = ", n_hid, ")"

        W_in_values = np.asarray(initWeights((n_in, n_hid), scale='glorot', mode='uni'), dtype=theano.config.floatX)
        self.W_in = theano.shared(W_in_values, name='W_in', borrow=True)
        W_hid_values = np.asarray(initWeights((n_hid, n_hid), mode='rnn'), dtype=theano.config.floatX)
        self.W_hid = theano.shared(W_hid_values, name='W_hid', borrow=True)
        b_hid_values = np.asarray(initWeights((n_hid, ), scale=1e-6, mode='fix-uni'), dtype=theano.config.floatX)
        self.b_hid = theano.shared(b_hid_values, name='b_hid', borrow=True)
        hid_0_values = np.zeros(n_hid, dtype=theano.config.floatX)
        self.hid_0 = theano.shared(hid_0_values, name='hid_0', borrow=True)

        W_in, W_hid, b_hid, hid_0 = self.W_in, self.W_hid, self.b_hid, self.hid_0
        self.params = [W_in, W_hid, b_hid, hid_0]

        # Select non-linearities
        if activation_func == 'tanh':  # range = [-1,1]
            act = T.tanh  # shape: (batch_size, num_outputs)
        elif activation_func == 'relu':  # rectified linear unit ,range = [0,inf]
            act = lambda x: x * (x > 0)  #T.maximum(lin_output,T.zeros_like(lin_output))
        elif activation_func == 'abs':  # abs unit ,range = [0,inf]
            act = T.abs_
        elif activation_func == 'sigmoid':  # range = [0,1]
            print "WARNING: sig() used!"
            #print "WARNING: consider using tanh(.) or relu(.) instead! Sigmoid is really BAD! (relu > tanh >> sigmoid)"
            act = T.nnet.sigmoid  #1/(1 + T.exp(-lin_output))
        elif activation_func == 'linear':
            print "Warning: linear activation in recurrent layer with fanout-%i! Is this the output layer?" % n_hid
            act = lambda x: x
        else:
            raise NotImplementedError("options are: activation_func=('relu'|'sigmoid'|'tanh'|'abs')")

        def recurrence(x_t, hid_prev):
            hid_lin_t = T.dot(x_t, W_in) + T.dot(hid_prev, W_hid) + b_hid
            hid_t = act(hid_lin_t)
            return [hid_t, hid_lin_t]

        outputs_info = [dict(initial=T.alloc(hid_0, input.shape[1], n_hid), taps=[-1]), dict()]
        # shapes are [time, batch, feat]
        ([hid, hid_lin], updates) = theano.scan(fn=recurrence,
                                                sequences=input,
                                                outputs_info=outputs_info,
                                                name='Recurrence')
        hid_lin = hid_lin.dimshuffle(1, 0, 2)  # exchange batch and time  again --> [batch, time, hid/feat]
        hid = act(hid_lin)  # I think this is needed for structural damping (calculating grad wrt hid_lin)

        self.output = hid[:, -1]  # [batch, hid/feat]
        self.hid = hid
        self.hid_lin = hid_lin

    def randomizeWeights(self, scale_w=1.0):
        n_in, n_hid = self.n_in, self.n_hid
        W_in_values = np.asarray(initWeights((n_in, n_hid), scale='glorot', mode='uni'), dtype=theano.config.floatX)
        self.W_in.set_value(W_in_values)
        W_hid_values = np.asarray(initWeights((n_in, n_hid), mode='rnn'), dtype=theano.config.floatX)
        self.W_hid.set_value(W_hid_values)
        b_hid_values = np.asarray(initWeights((n_hid, ), scale=1e-6, mode='fix-uni'), dtype=theano.config.floatX)
        self.b_hid.set_value(b_hid_values)
        hid_0_values = np.zeros(n_hid, dtype=theano.config.floatX)
        self.hid_0.set_value(hid_0_values)
